#! /usr/bin/python3
#! -*- coding:utf-8; mode:python; -*-

__all__ = [ ]

import glob
import logging
import math
import os
import shutil
import stat
import sys

from .cli import *
from .configuration import *
from .constants import *
from .errors import *
from .manager import *
from .utility import *

PROGRAM_NAME = 'integrity'

_REQUIRED_FOR_GRAPH = [
    'id',
    'generator',
    'file',
    'nodes',
    'edges',
]

_REQUIRED_FOR_LAYOUT = [
    'id',
    'graph',
    'file',
]

_REQUIRED_FOR_PROPERTY_SCALED_HISTOGRAM = [
    'layout',
    'size',
    'minimum',
    'maximum',
    'mean',
    'rms',
]

_REQUIRED_FOR_PROPERTY_GLIDING_AVERAGE = [
    'layout',
    'file',
    'vicinity',
    'size',
    'minimum',
    'maximum',
    'mean',
    'rms',
    'sigma',
    'points',
    'max-x',
    'max-y',
]

_REQUIRED_FOR_INTER_LAYOUTS = [
    'id',
    'parent1st',
    'parent2nd',
    'method',
    'rate',
]

_REQUIRED_FOR_WORSE_LAYOUTS = [
    'id',
    'parent',
    'method',
    'rate',
]

class Checker(object):

    def __init__(self, mngr, repair=False):
        self.__mngr = mngr
        self.__repair = repair
        self.__issues = None

    def __call__(self):
        self.__mngr.sql_exec("PRAGMA foreign_keys = OFF", tuple())
        self.__issues = 0
        self.__check_tempdirs()
        self.__check_graphs()
        self.__check_layouts()
        self.__check_properties()
        self.__check_histograms()
        self.__check_princomp()

    @property
    def issues(self):
        return self.__issues

    def __bemoan(self, message):
        self.__issues += 1
        logging.warning(message)

    def __remove_from_db(self, curs, query, *params):
        logging.debug(query.replace('?', '%s') % tuple(sqlrepr(p) for p in params))
        if self.__repair:
            self.__mngr.sql_exec_curs(curs, query, params)

    def __remove_from_fs(self, filename):
        logging.debug("Removing file {!r}".format(filename))
        if self.__repair:
            try: os.remove(filename)
            except FileNotFoundError: pass

    def __check_tempdirs(self):
        for tempdir in glob.glob('temp-*'):
            if os.path.isdir(tempdir):
                self.__bemoan("Orphaned temporary directory {!r}".format(tempdir))
                if self.__repair:
                    shutil.rmtree(tempdir)
            else:
                self.__bemoan("Orphaned temporary file {!r}".format(tempdir))
                self.__remove_from_fs(tempdir)

    def __check_graphs(self):
        self.__check_table('Graphs', _REQUIRED_FOR_GRAPH)
        self.__check_files('Graphs', os.path.join(glob.escape(self.__mngr.graphsdir), '*'))

    def __check_layouts(self):
        self.__check_table('Layouts', _REQUIRED_FOR_LAYOUT, crossrefs={ 'graph' : ('Graphs', 'id') })
        self.__check_files('Layouts', os.path.join(glob.escape(self.__mngr.layoutdir), '*', '*'))
        self.__check_inter_layouts()
        self.__check_worse_layouts()
        with self.__mngr.sql_ctx as curs:
            for row in self.__mngr.sql_select_curs(curs, 'Layouts', layout=None):
                layoutid = row['id']
                filename = row['file']
                if self.__mngr.sql_select_curs(curs, 'InterLayouts', id=layoutid):
                    continue
                if self.__mngr.sql_select_curs(curs, 'WorseLayouts', id=layoutid):
                    continue
                self.__bemoan("Improper layout {!s} not mentioned for any transformation".format(layoutid))
                self.__remove_from_db(curs, "DELETE FROM `Layouts` WHERE `id` = ?", layoutid)
                self.__remove_from_fs(filename)

    def __check_inter_layouts(self):
        crossrefs = { 'id' : ('Layouts', 'id'), 'parent1st' : ('Layouts', 'id'), 'parent2nd' : ('Layouts', 'id') }
        self.__check_table('InterLayouts', _REQUIRED_FOR_INTER_LAYOUTS, crossrefs=crossrefs)

    def __check_worse_layouts(self):
        crossrefs = { 'id' : ('Layouts', 'id'), 'parent' : ('Layouts', 'id') }
        self.__check_table('WorseLayouts', _REQUIRED_FOR_WORSE_LAYOUTS, crossrefs=crossrefs)

    def __check_properties(self):
        for prop in Properties:
            kind = PropertyKinds.classify(prop)
            table = SQL_PROPERTY_TABLES[prop]
            pattern = os.path.join(glob.escape(self.__mngr.propsdir), '**', glob.escape(enum_to_json(prop)), '*')
            crossrefs = { 'layout' : ('Layouts', 'id') }
            if kind is PropertyKinds.SCALED_HISTOGRAM:
                self.__check_table(table, _REQUIRED_FOR_PROPERTY_SCALED_HISTOGRAM, crossrefs=crossrefs)
                self.__check_files('ScaledHistograms', pattern, recursive=True, where={'property' : prop})
            elif kind is PropertyKinds.GLIDING_AVERAGE:
                self.__check_table(table, _REQUIRED_FOR_PROPERTY_GLIDING_AVERAGE, crossrefs=crossrefs)
                self.__check_files(table, pattern)
            else:
                raise ValueError(kind)

    def __check_histograms(self):
        table = 'ScaledHistograms'
        required = [ 'property', 'layout', 'binning', 'bincount', 'binwidth', 'entropy' ]
        crossrefs = { 'layout' : ('Layouts', 'id') }
        self.__check_table('ScaledHistograms', required=required, crossrefs=crossrefs)
        with self.__mngr.sql_ctx as curs:
            for row in self.__mngr.sql_select_curs(curs, 'ScaledHistograms'):
                layoutid = row['layout']
                prop = row['property']
                proptab = SQL_PROPERTY_TABLES[prop]
                if not self.__mngr.sql_select_curs(curs, proptab, layout=layoutid):
                    self.__bemoan("Layout {!s} mentioned in table {!r} but not {!r}".format(layoutid, table, proptab))
                    self.__remove_from_db(curs, "DELETE FROM `{:s}` WHERE `layout` = ?".format(table), layoutid)

    def __check_princomp(self):
        with self.__mngr.sql_ctx as curs:
            for layoutid in map(lambda r : r['id'], self.__mngr.sql_select_curs(curs, 'Layouts')):
                getter = lambda r : (r['component@0'], r['component@1'])
                vecabs = lambda v : math.sqrt(sum(x**2 for x in v))
                vecdot = lambda v1, v2 : sum(x1 * x2 for (x1, x2) in zip(v1, v2))
                major = get_one_or(map(getter, self.__mngr.sql_select_curs(curs, 'PrinComp1st', layout=layoutid)))
                minor = get_one_or(map(getter, self.__mngr.sql_select_curs(curs, 'PrinComp2nd', layout=layoutid)))
                delete = set()
                if major is not None:
                    if vecabs(major) != 0.0 and abs(1.0 - vecabs(major)) > 1.0E-3:
                        self.__bemoan("Major component of layout {!s} is not normalized".format(layoutid))
                        delete.add('PrinComp1st')
                if minor is not None:
                    if vecabs(minor) != 0.0 and abs(1.0 - vecabs(minor)) > 1.0E-3:
                        self.__bemoan("Minor component of layout {!s} is not normalized".format(layoutid))
                        delete.add('PrinComp2nd')
                if major is not None and minor is not None and abs(vecdot(major, minor)) > 1.0E-3:
                    self.__bemoan("Principial components of layout {!s} are not orthogonal".format(layoutid))
                    delete.add('PrinComp1st')
                    delete.add('PrinComp2nd')
                for table in delete:
                    self.__remove_from_db(curs, "DELETE FROM `{:s}` WHERE `layout` = ?".format(table), layoutid)

    def __check_table(self, table, required, crossrefs=None):
        logging.info("Checking database table {!r} ...".format(table))
        with self.__mngr.sql_ctx as curs:
            nullcount = 0
            for row in self.__mngr.sql_select_curs(curs, table):
                for col in required:
                    if row[col] is None:
                        nullcount += 1
                for col in value_or(crossrefs, default=dict()).keys():
                    (reftab, refcol) = crossrefs[col]
                    if not self.__mngr.sql_select_curs(curs, reftab, **{ refcol : row[col] }):
                        self.__bemoan("Dangling reference from {:s}.{:s} to {:s}.{:s} on key {!r}".format(
                            table, col, reftab, refcol, row[col]))
                        self.__remove_from_db(curs, "DELETE FROM `{:s}` WHERE `{:s}` = ?".format(table, col), row[col])
            if nullcount:
                self.__bemoan("{:,d} required values are missing from table {!r}".format(nullcount, table))
            for col in required:
                self.__remove_from_db(curs, "DELETE FROM `{:s}` WHERE `{:s}` ISNULL".format(table, col))

    def __check_files(self, table, pattern, column='file', where=None, recursive=False):
        if where is None: where = dict()
        with self.__mngr.sql_ctx as curs:
            files_in_db = set()
            for row in self.__mngr.sql_select_curs(curs, table, **where):
                filename = row[column]
                files_in_db.add(filename)
                if not os.path.isfile(filename):
                    self.__bemoan("File does not exist: {:s}".format(filename))
                    self.__remove_from_db(curs, "DELETE FROM `{:s}` WHERE `{:s}` = ?".format(table, column), filename)
            files_on_disk = set(glob.glob(pattern, recursive=recursive))
            logging.info("{:,d} files match {!r}".format(len(files_on_disk), pattern))
            logging.info("{:,d} files referenced in {:s}.{:s}".format(len(files_in_db), table, column))
            for filename in files_on_disk - files_in_db:
                (what, size) = _get_file_info(filename)
                self.__bemoan("Stray file {!r} ({:s}, {:,d} bytes)".format(filename, what, size))
                self.__remove_from_fs(filename)

def _get_file_info(filename):
    st = os.stat(filename, follow_symlinks=False)
    if stat.S_ISSOCK(st.st_mode):
        what = "socket"
    elif stat.S_ISLNK(st.st_mode):
        what = "symbolic link" if os.path.exists(filename) else "broken symbolic link"
    elif stat.S_ISREG(st.st_mode):
        what = "regular file"
    elif stat.S_ISBLK(st.st_mode):
        what = "block special device"
    elif stat.S_ISDIR(st.st_mode):
        what = "directory"
    elif stat.S_ISCHR(st.st_mode):
        what = "character special device"
    elif stat.S_ISIFO(st.st_mode):
        what = "FIFO"
    elif stat.S_ISDOOR(st.st_mode):
        what = "door"
    elif stat.S_ISPORT(st.st_mode):
        what = "event port"
    elif stat.S_ISWHT(st.st_mode):
        what = "whiteout"
    else:
        what = "unknown"
    return (what, st.st_size)

class Main(AbstractMain):

    def _run(self, ns):
        config = Configuration(ns.configdir)
        with Manager(absexecdir=ns.execdir, absdatadir=ns.datadir, config=config) as mngr:
            chkr = Checker(mngr, ns.repair)
            chkr()
            if chkr.issues:
                raise SanityError("{:d} issues found".format(chkr.issues))

    def _argparse_hook_before(self, ap):
        ap.add_argument('--repair', '-r', action='store_true', help="Attempt to fix any detected issues")

if __name__ == '__main__':
    kwargs = {
        'prog' : PROGRAM_NAME,
        'usage' : "%(prog)s [--repair]",
        'description' : "Performs some integrity checks on the database and file system storage.",
    }
    with Main(**kwargs) as app:
        app(sys.argv[1 : ])
