#! /usr/bin/python3
#! -*- coding:utf-8; mode:python; -*-

__all__ = [ 'do_layouts' ]

import itertools
import logging
import os

from .configuration import *
from .constants import *
from .errors import *
from .manager import *
from .utility import *

_LAYOUT_PROGS = {
    Layouts.NATIVE         : None,
    Layouts.FMMM           : 'force',
    Layouts.STRESS         : 'force',
    Layouts.RANDOM_UNIFORM : 'random',
    Layouts.RANDOM_NORMAL  : 'random',
    Layouts.PHANTOM        : 'phantom',
}

_LAYOUT_FLAGS = {
    Layouts.NATIVE         : None,
    Layouts.FMMM           : [ '--algorithm=FMMM' ],
    Layouts.STRESS         : [ '--algorithm=STRESS' ],
    Layouts.RANDOM_UNIFORM : [ '--distribution=UNIFORM' ],
    Layouts.RANDOM_NORMAL  : [ '--distribution=NORMAL' ],
    Layouts.PHANTOM        : [ ],
}

assert all(l in _LAYOUT_PROGS for l in Layouts)
assert all(l in _LAYOUT_FLAGS for l in Layouts)

def do_layouts(mngr : Manager, badlog : BadLog):
    for row in mngr.sql_select('Graphs'):
        graphid = row['id']
        graphfilename = row['file']
        size = GraphSizes.classify(row['nodes'])
        for layout in mngr.config.desired_layouts.keys():
            if size in mngr.config.desired_layouts[layout] and not mngr.sql_select('Layouts', graph=graphid, layout=layout):
                if badlog.get_bad(Actions.LAYOUTS, graphid, layout):
                    logging.notice("Skipping {:s} layout for graph {!s}".format(layout.name, graphid))
                    continue
                try:
                    if layout is Layouts.NATIVE:
                        if row['native']:
                            _lay_native(mngr, graphid, graphfilename)
                    else:
                        _lay_generic(mngr, layout, graphid, graphfilename=graphfilename)
                except RecoverableError as e:
                    badlog.set_bad(Actions.LAYOUTS, graphid, layout)
                    logging.error("Cannot make {:s} layout for graph {!s}: {!s}".format(layout.name, graphid, e))

def _lay_native(mngr: Manager, graphid : Id, graphfilename : str):
    with mngr.sql_ctx as curs:
        layoutid = mngr.make_unique_layout_id(curs)
        filename = mngr.make_layout_filename(graphid, layoutid, layout=Layouts.NATIVE)
        mngr.sql_insert_curs(curs, 'Layouts', id=layoutid, graph=graphid, layout=Layouts.NATIVE, file=filename)
    directory = os.path.dirname(filename)
    os.makedirs(directory, exist_ok=True)
    symlinktarget = os.path.relpath(graphfilename, start=directory)
    logging.info("Creating symbolic link: {!r} -> {!r}".format(filename, symlinktarget))
    os.symlink(symlinktarget, filename)

def _lay_generic(mngr : Manager, layout : Layouts, graphid : Id, graphfilename : str):
    assert layout is not Layouts.NATIVE
    logging.info("Generating {:s} layout for graph {!s} ...".format(layout.name, graphid))
    with mngr.make_tempdir() as tempdir:
        cmd = [ os.path.join(mngr.abs_execdir, 'src', 'layouts', _LAYOUT_PROGS[layout]) ]
        cmd.append('--output={:s}'.format(os.path.join(tempdir, enum_to_json(layout) + LAYOUT_FILE_SUFFIX)))
        cmd.append('--meta={:s}'.format('STDIO'))
        cmd.extend(_LAYOUT_FLAGS[layout])
        cmd.append(graphfilename)
        meta = mngr.call_graphstudy_tool(cmd, meta='stdout')
        with mngr.sql_ctx as curs:
            layoutid = mngr.make_unique_layout_id(curs)
            layoutfilename = mngr.make_layout_filename(graphid, layoutid, layout)
            mngr.sql_insert_curs(
                curs, 'Layouts', id=layoutid, graph=graphid, layout=layout, file=layoutfilename,
                width=meta.get('width'), height=meta.get('height'), seed=encoded(meta.get('seed')),
            )
        logging.debug("Renaming layout file {!r} to {!r}".format(meta['filename'], layoutfilename))
        os.makedirs(os.path.dirname(layoutfilename), exist_ok=True)
        os.rename(meta['filename'], layoutfilename)
