#! /usr/bin/python3
#! -*- coding:utf-8; mode:python; -*-

__all__ = [ 'do_model', 'Oracle' ]

import collections
import itertools
import logging
import math
import os
import random
import statistics
import time

import numpy as np
_keras = None

from .configuration import *
from .constants import *
from .errors import *
from .manager import *
from .nn import *
from .quarry import *
from .utility import *

_MIN_SIGNIFICANCE = 0.15

_FP_TYPE = 'float32'

_DEBUG_DIR_ENVVAR = 'MSC_MODEL_DEBUGDIR'

_LAYOUT_RATINGS = {
    Layouts.NATIVE         : +1.0,
    Layouts.FMMM           : +1.0,
    Layouts.STRESS         : +1.0,
    Layouts.PHANTOM        : -1.0,
    Layouts.RANDOM_NORMAL  : -1.0,
    Layouts.RANDOM_UNIFORM : -1.0,
}

_INVALID_TIME_STAMP = math.nan

_MeanStdev = collections.namedtuple('MeanStdev', [ 'mean', 'stdev' ])

def _import_3d_party_libraries():
    global _keras
    import keras
    _keras = keras

def _log_discard_too_old(filename, timestamp, noolder, what="stuff"):
    limit = datetime.fromtimestamp(noolder) if math.isfinite(noolder) else "now"
    logging.notice("Discarding NN {:s} loaded from {!r} with time-stamp {!s} older than {!s}".format(
        what, filename, timestamp, limit))

class _DataSet(object):

    def __init__(self, lhs, rhs, aux, out):
        assert [ len(lhs.shape), len(rhs.shape), len(aux.shape), len(out.shape) ] == [ 2, 2, 2, 2 ]
        assert lhs.shape[0] == rhs.shape[0] == aux.shape[0] == out.shape[0]
        assert lhs.shape[1] == rhs.shape[1]
        assert out.shape[1] == 1
        self.lhs = lhs
        self.rhs = rhs
        self.aux = aux
        self.out = out

    def __len__(self):
        return self.out.shape[0]

    def bias(self):
        pos = (self.out > 0.0).sum()
        neg = (self.out < 0.0).sum()
        bias = self.out.mean()
        return { 'pos' : pos, 'neg' : neg, 'bia' : bias }

def _save_weighted_model(mngr, model):
    logging.info("Saving NN model architecture to file {!r} ...".format(mngr.nn_model))
    with open(mngr.nn_model, 'w') as ostr:
        ostr.write(model.to_yaml())
    logging.info("Saving NN model weights to file {!r} ...".format(mngr.nn_weights))
    model.save_weights(mngr.nn_weights)

def _restore_weighted_model(mngr):
    logging.info("Trying to load NN model architecture from file {!r} ...".format(mngr.nn_model))
    try:
        with open(mngr.nn_model, 'r') as ostr:
            yaml = ostr.read()
    except FileNotFoundError:
        logging.error("Cannot load NN model architecture")
        return None
    model = _keras.models.model_from_yaml(yaml)
    logging.info("Trying to load NN model weights from file {!r} ...".format(mngr.nn_weights))
    try:
        model.load_weights(mngr.nn_weights)
    except OSError:  # For whatever reason, no FileNotFoundError but some other OSError is raised here.
        logging.error("Cannot load NN model weights")
        return None
    return model

def _save_features(mngr, layout, graph, exception=FatalError):
    logging.info("Saving NN features to file {!r} ...".format(mngr.nn_features))
    lf = _layout_features(mngr, exception=exception)
    gf = _graph_features(mngr, exception=exception)
    features = dict()
    for (feat, of, norm) in [ (lf, 'layout', layout), (gf, 'graph', graph) ]:
        vector = features[of] = list()
        for (name, (mean, stdev)) in zip(feat, norm):
            vector.append(NNFeature(float(mean), float(stdev), name=name))
    with open(mngr.nn_features, 'wb') as ostr:
        pickle_objects(ostr, features)

def _restore_features(mngr, noolder=None):
    logging.info("Trying to restore NN features from file {!r} ...".format(mngr.nn_features))
    try:
        with open(mngr.nn_features, 'rb') as istr:
            timestamp, features = unpickle_objects(istr, dict)
    except FileNotFoundError:
        return _INVALID_TIME_STAMP, None, None
    if noolder is None or timestamp.timestamp() >= noolder:
        layout = np.array([ [ feat.mean, feat.stdev ] for feat in features['layout'] ], dtype=_FP_TYPE).reshape(-1, 2)
        graph  = np.array([ [ feat.mean, feat.stdev ] for feat in features['graph']  ], dtype=_FP_TYPE).reshape(-1, 2)
        return timestamp.timestamp(), layout, graph
    else:
        _log_discard_too_old(mngr.nn_features, timestamp, noolder, what="features")
        return _INVALID_TIME_STAMP, None, None

def _save_test_score(mngr, info, expected, actual):
    logging.info("Saving NN test score ...")
    assert expected.shape == (len(info), 1)
    assert actual.shape == (len(info), 1)
    results = list()
    for (i, (lhs, rhs)) in enumerate(info):
        exp = float(expected[i][0])
        act = float(actual[i][0])
        results.append((lhs, rhs, exp, act))
    with mngr.sql_ctx as curs:
        curs.execute("DELETE FROM `NNTestScore`", tuple())
        curs.executemany("INSERT INTO `NNTestScore` VALUES (?, ?, ?, ?)", iter(results))

def _restore_testscore(mngr, noolder=None):
    raise NotImplementedError("YAGNI")

def _get_layout_features_curs(mngr, curs, layoutid, cache=None):
    try:
        return _cache_get(cache, layoutid)
    except KeyError:
        pass
    for row in mngr.sql_select_curs(curs, 'LayoutFeatures', layout=layoutid):
        return _cache_put(cache, layoutid, row[1:])
    raise KeyError("Layout {!s} does not exist".format(layoutid))

def _get_graph_features_curs(mngr, curs, graphid, cache=None):
    try:
        return _cache_get(cache, graphid)
    except KeyError:
        pass
    for row in mngr.sql_select_curs(curs, 'GraphFeatures', graph=graphid):
        return _cache_put(cache, graphid, row[1:])
    raise KeyError("Graph {!s} does not exist".format(graphid))

def _cache_get(cache, key):
    if cache is None:
        raise KeyError()
    return cache[key]

def _cache_put(cache, key, value):
    if cache is not None:
        cache[key] = value
    return value

def _do_compile_model(mngr, model):
    model.compile(loss='mse', optimizer='sgd')

def _layout_features(mngr, exception=FatalError):
    return _get_sql_view_columns(mngr, 'LayoutFeatures', exception=exception)

def _graph_features(mngr, exception=FatalError):
    return _get_sql_view_columns(mngr, 'GraphFeatures', exception=exception)

def _get_sql_view_columns(mngr, view, exception=FatalError):
    info = mngr.sql_exec("PRAGMA table_info(`{:s}`)".format(view), tuple())
    if len(info) <= 1:
        raise SanityError("The view {!r} seems to contain no data at all".format(view))
    return [ info[i]['name'] for i in range(1, len(info)) ]

class Oracle(object):

    def __init__(self, mngr, exception=FatalError):
        _import_3d_party_libraries()
        self.__manager = mngr
        self.__model = _restore_weighted_model(self.__manager)
        if self.__model is None:
            raise exception("There is no NN model to be used")
        timestamp, self.__lonorm, self.__grnorm = _restore_features(self.__manager)
        if self.__lonorm is None or self.__grnorm is None:
            raise exception("There are no NN freatures to be used")

    def __call__(self, layoutpairs, exception=FatalError, bidirectional=False):
        n = len(layoutpairs)
        forward = self.__load_data(layoutpairs, exception=exception)
        self.__predict(forward)
        if not bidirectional:
            return list(float(forward.out[i][0]) for i in range(n))
        backward = _DataSet(forward.rhs, forward.lhs, forward.aux, np.zeros(n, dtype=_FP_TYPE).reshape(n, 1))
        self.__predict(backward)
        return list((float(forward.out[i][0]), float(backward.out[i][0])) for i in range(n))

    def __load_data(self, layoutpairs, exception=FatalError):
        n = len(layoutpairs)
        logging.info("Loading data for {:,d} layout pairs from database ...".format(n))
        t0 = time.time()
        lhs = list()
        rhs = list()
        aux = list()
        with self.__manager.sql_ctx as curs:
            for (lhslid, rhslid) in layoutpairs:
                graphid = _get_same_graph_id_curs(self.__manager, curs, lhslid, rhslid, exception=exception)
                lhs.append(_get_layout_features_curs(self.__manager, curs, lhslid))
                rhs.append(_get_layout_features_curs(self.__manager, curs, rhslid))
                aux.append(_get_graph_features_curs(self.__manager, curs, graphid))
        lhs = _normalize_data(self.__lonorm, np.array(lhs, dtype=_FP_TYPE).reshape(n, self.__lonorm.shape[0]))
        rhs = _normalize_data(self.__lonorm, np.array(rhs, dtype=_FP_TYPE).reshape(n, self.__lonorm.shape[0]))
        aux = _normalize_data(self.__grnorm, np.array(aux, dtype=_FP_TYPE).reshape(n, self.__grnorm.shape[0]))
        out = np.zeros(n, dtype=_FP_TYPE).reshape((n, 1))  # TODO: Not the smartest move ...
        t1 = time.time()
        logging.info("Data acquisition took {:.3f} seconds".format(t1 - t0))
        return _DataSet(lhs, rhs, aux, out)

    def __predict(self, data):
        logging.info("Running NN model to classify {:,d} layout pairs ...".format(len(data)))
        t0 = time.time()
        prediction = self.__model.predict([ data.lhs, data.rhs, data.aux ])
        # For whatever reason, predict() returns no Numpy array but a Python list if the array is empty...
        if len(prediction) > 0:
            assert data.out.shape == prediction.shape
            np.copyto(data.out, prediction)
        t1 = time.time()
        logging.info("Classification took {:.3f} seconds".format(t1 - t0))
        return data.out

def _get_graph_id_curs(mngr, curs, layoutid, exception=FatalError):
    row = get_one_or(mngr.sql_select_curs(curs, 'Layouts', id=layoutid))
    if row is None:
        raise exception("The layout {!s} does not exist".format(layoutid))
    return row['graph']

def _get_same_graph_id_curs(mngr, curs, lhslid, rhslid, exception=FatalError):
    lhsgid = _get_graph_id_curs(mngr, curs, lhslid, exception=exception)
    rhsgid = _get_graph_id_curs(mngr, curs, rhslid, exception=exception)
    if lhsgid != rhsgid:
        logging.notice("Layout {!s} corresponds to graph {!s}".format(lhslid, lhsgid))
        logging.notice("Layout {!s} corresponds to graph {!s}".format(rhslid, rhsgid))
        raise exception("The layouts {!s} and {!s} correspond to different graphs".format(lhslid, rhslid))
    return lhsgid

def _get_debugdir():
    directory = os.getenv(_DEBUG_DIR_ENVVAR)
    if directory is None:
        return None
    if not os.path.isabs(directory):
        logging.warning("Ignoring non-absolute directory {:s}={!r}".format(_DEBUG_DIR_ENVVAR, directory))
        return None
    if not os.path.isdir(directory):
        logging.warning("Ignoring non-existant directory {:s}={!r}".format(_DEBUG_DIR_ENVVAR, directory))
        return None
    return directory

def do_model(mngr : Manager, badlog : BadLog):
    _import_3d_party_libraries()
    os.makedirs(mngr.nndir, exist_ok=True)
    lf = _layout_features(mngr)
    gf = _graph_features(mngr)
    model = _build_model(mngr, inputs=len(lf), auxinputs=len(gf))
    (training, testing, testinfo) = _load_training_and_testing_data(mngr)
    _train_model(model, training)
    predictions = _test_model(model, testing)
    _save_weighted_model(mngr, model)
    _save_test_score(mngr, testinfo, expected=testing.out, actual=predictions)

def _load_training_and_testing_data(mngr : Manager):
    logging.info("Loading training / testing data from database ...")
    t0 = time.time()
    grdata = dict()
    lodata = dict()
    lhs = list()
    rhs = list()
    aux = list()
    out = list()
    info = list()
    with mngr.sql_ctx as curs:
        datafilter = lambda seq : filter(lambda t : abs(t[3]) >= _MIN_SIGNIFICANCE, seq)
        collection = list()
        collection.extend(datafilter(_gather_proper(mngr, curs)))
        collection.extend(datafilter(_gather_inter(mngr, curs)))
        collection.extend(datafilter(_gather_worse(mngr, curs)))
        random.shuffle(collection)
        for (lhsid, rhsid, graphid, rank) in collection:
            lhs.append(_get_layout_features_curs(mngr, curs, lhsid, cache=lodata))
            rhs.append(_get_layout_features_curs(mngr, curs, rhsid, cache=lodata))
            aux.append(_get_graph_features_curs(mngr, curs, graphid, cache=grdata))
            out.append([ rank ])
            info.append((lhsid, rhsid))
    count = len(out)
    if count < 10:
        raise FatalError("There is no point in training a neural network with only {:,d} data points".format(count))
    grdata = np.array(list(grdata.values()), dtype=_FP_TYPE)
    lodata = np.array(list(lodata.values()), dtype=_FP_TYPE)
    grnorm = _get_normalizers(grdata)
    lonorm = _get_normalizers(lodata)
    lhs = _normalize_data(lonorm, np.array(lhs, dtype=_FP_TYPE))
    rhs = _normalize_data(lonorm, np.array(rhs, dtype=_FP_TYPE))
    aux = _normalize_data(grnorm, np.array(aux, dtype=_FP_TYPE))
    out = np.array(out, dtype=_FP_TYPE)
    selection = (np.random.rand(count) > 0.2)
    cnt1 = selection.sum()
    cnt2 = count - cnt1
    logging.info("Loaded corpus with {:,d} events".format(count))
    logging.info("Using {:,d} events ({:7.3f} %) for training".format(cnt1, 100.0 * cnt1 / count))
    logging.info("Using {:,d} events ({:7.3f} %) for testing".format(cnt2, 100.0 * cnt2 / count))
    training = _DataSet(lhs[ selection], rhs[ selection], aux[ selection], out[ selection])
    testing  = _DataSet(lhs[~selection], rhs[~selection], aux[~selection], out[~selection])
    logging.info("Training set has {pos:,d} / {neg:,d} distribution (bias = {bia:.5f})".format(**training.bias()))
    logging.info("Testing set has {pos:,d} / {neg:,d} distribution (bias = {bia:.5f})".format(**testing.bias()))
    testinfo = [ info[i] for i in range(count) if not selection[i] ]
    t1 = time.time()
    logging.info("Data acquisition took {:.3f} seconds".format(t1 - t0))
    _save_features(mngr, layout=lonorm, graph=grnorm)
    return (training, testing, testinfo)

def _gather_proper(mngr, curs):
    for grow in mngr.sql_select_curs(curs, 'Graphs'):
        graphid = grow['id']
        layouts = {
            r['id'] : r['layout'] for r in mngr.sql_select_curs(curs, 'Layouts', graph=graphid, layout=object)
        }
        for (lhs, rhs) in itertools.combinations(list(layouts.keys()), 2):
            if random.random() > 0.5:
                (lhs, rhs) = (rhs, lhs)
            rank = _rank_layouts(layouts[lhs], layouts[rhs])
            if rank is not None:
                assert -1.0 <= rank <= +1.0
                yield (lhs, rhs, graphid, rank)

def _gather_inter(mngr, curs):
    for method in LayInter:
        graphs = dict()
        lines = collections.defaultdict(dict)
        for row in mngr.sql_select_curs(curs, 'InterLayouts', method=method):
            parents = (row['parent1st'], row['parent2nd'])
            graphs[parents] = _get_graph_id(mngr, curs, row['id'])
            lines[parents][row['rate']] = row['id']
        for parents in lines.keys():
            lines[parents][0.0] = parents[0]
            lines[parents][1.0] = parents[1]
        for parents in graphs.keys():
            graphid = graphs[parents]
            prank = _rank_layout_ids(mngr, curs, parents[0], parents[1])
            if prank is None:
                continue
            for (r1, r2) in itertools.combinations(lines[parents].keys(), 2):
                if random.random() > 0.5:
                    (r1, r2) = (r2, r1)
                rank = (r2 - r1) * prank
                assert -1.0 <= rank <= +1.0
                yield (lines[parents][r1], lines[parents][r2], graphid, rank)

def _gather_worse(mngr, curs):
    maxima = _get_worse_max(mngr, curs)
    for method in LayWorse:
        graphs = dict()
        lines = collections.defaultdict(dict)
        for row in mngr.sql_select_curs(curs, 'WorseLayouts', method=method):
            parent = row['parent']
            graphs[parent] = _get_graph_id(mngr, curs, row['id'])
            lines[parent][row['rate']] = row['id']
        for parent in lines.keys():
            lines[parent][0.0] = parent
        for (parent, graphid) in graphs.items():
            prank = _rank_layout_id(mngr, curs, parent)
            if prank is None:
                continue
            for (r1, r2) in itertools.combinations(lines[parent].keys(), 2):
                if random.random() > 0.5:
                    (r1, r2) = (r2, r1)
                r1norm = r1 / maxima[method]
                r2norm = r2 / maxima[method]
                rank = (r1norm - r2norm) * prank if prank > 0.0 else 0.0
                assert -1.0 <= rank <= +1.0
                yield (lines[parent][r1], lines[parent][r2], graphid, rank)

def _get_graph_id(mngr, curs, layoutid : Id) -> Id:
    return get_one(mngr.sql_select_curs(curs, 'Layouts', id=layoutid))['graph']

def _rank_layout_id(mngr, curs, layoutid : Id) -> float:
    return _rank_layout(_get_layout(mngr, curs, layoutid))

def _rank_layout_ids(mngr, curs, lhsid : Id, rhsid : Id) -> float:
    return _rank_layouts(_get_layout(mngr, curs, lhsid), _get_layout(mngr, curs, rhsid))

def _get_layout(mngr, curs, layoutid : Id) -> Layouts:
    return get_one(mngr.sql_select_curs(curs, 'Layouts', id=layoutid))['layout']

def _build_model(mngr, inputs : int, auxinputs : int):
    logging.info("Constructing NN model ...")
    t0 = time.time()
    kwargs = {
        'kernel_initializer' : 'truncated_normal',
        'bias_initializer'   : 'zeros',
        'dtype'              : _FP_TYPE,
    }
    lhsin = _keras.layers.Input(shape=singleton(inputs), dtype=_FP_TYPE, name='lhsin')
    rhsin = _keras.layers.Input(shape=singleton(inputs), dtype=_FP_TYPE, name='rhsin')
    auxin = _keras.layers.Input(shape=singleton(auxinputs), dtype=_FP_TYPE, name='auxin')
    nn = _make_half_model(inputs, kwargs)
    lhs = nn(lhsin)
    rhs = nn(rhsin)
    sub = _keras.layers.Subtract(dtype=_FP_TYPE, name='sub')([ lhs, rhs ])
    aux = _keras.layers.Dense(auxinputs, **kwargs, activation='linear', name='aux')(auxin)
    cat = _keras.layers.Concatenate(dtype=_FP_TYPE, name='cat')([ sub, aux ])
    out = _keras.layers.Dense(1, **kwargs, activation='tanh', name='out')(cat)
    model = _keras.models.Model([ lhsin, rhsin, auxin ], out, name='total')
    debugdir = _get_debugdir()
    if debugdir is not None:
        logging.notice("Dumping auxiliary information (hint: {:s}={!r}) ...".format(_DEBUG_DIR_ENVVAR, debugdir))
        _dump_model_infos(model, nn, directory=debugdir)
    t1 = time.time()
    logging.info("Construction took {:.3f} seconds".format(t1 - t0))
    logging.info("Compiling NN model ...")
    t0 = time.time()
    _do_compile_model(mngr, model)
    t1 = time.time()
    logging.info("Compilation took {:.3f} seconds".format(t1 - t0))
    return model

def _make_half_model(inputs, kwargs, hidden_dims=10, output_dims=None):
    output_dims = value_or(output_dims, hidden_dims)
    layers = list()
    layers.append(_keras.layers.Input(shape=singleton(inputs), dtype=_FP_TYPE, name='in'))
    layers.append(_keras.layers.Dropout(0.50, dtype=_FP_TYPE, name='do1')(layers[-1]))
    layers.append(_keras.layers.Dense(hidden_dims, **kwargs, activation='linear', name='l1')(layers[-1]))
    layers.append(_keras.layers.Dropout(0.25, dtype=_FP_TYPE, name='do2')(layers[-1]))
    layers.append(_keras.layers.Dense(output_dims, **kwargs, activation='relu', name='l2')(layers[-1]))
    (il, *hidden, ol) = layers
    return _keras.models.Model(inputs=il, outputs=ol, name='shared')

def _train_model(model, data : _DataSet, epochs : int = 100):
    logging.info("Training NN model for {:,d} epochs ...".format(epochs))
    t0 = time.time()
    history = model.fit([ data.lhs, data.rhs, data.aux ], data.out, verbose=0, validation_split=0.25, epochs=epochs)
    t1 = time.time()
    logging.info("Training took {:.3f} seconds".format(t1 - t0))

def _test_model(model, data : _DataSet) -> np.array:
    logging.info("Testing NN model ...")
    t0 = time.time()
    predictions = model.predict([ data.lhs, data.rhs, data.aux ])
    assert predictions.shape == data.out.shape
    errors = predictions - data.out
    errmean = errors.mean()
    errstdev = errors.std()
    testsize = predictions.size
    hits = (predictions * data.out >= 0.0).sum()
    t1 = time.time()
    logging.info("Testing took {:.3f} seconds".format(t1 - t0))
    logging.info("Success rate: {:,d} of {:,d} ({:.3f} %)".format(hits, testsize, 100.0 * hits / testsize))
    logging.info("Prediction error: {:.3f} % +/- {:.3f} %".format(100.0 * errmean, 100.0 * errstdev))
    return predictions

def _get_worse_max(mngr, curs):
    return {
        r[0] : r[1] for r in mngr.sql_exec_curs(
            curs, "SELECT `method`, MAX(`rate`) FROM `WorseLayouts` GROUP BY `method`", tuple()
        )
    }

def _rank_layout(layout : Layouts) -> float:
    #return random.choice([-1.0, +1.0])
    assert layout is None or type(layout) is Layouts
    return _LAYOUT_RATINGS.get(layout)

def _rank_layouts(lhs : Layouts, rhs : Layouts) -> float:
    return (_rank_layout(rhs) - _rank_layout(lhs)) / 2.0

def _get_normalizers(data):
    assert len(data.shape) == 2
    n = data.shape[1]
    stats = np.zeros((n, 2), dtype=_FP_TYPE)
    for i in range(n):
        feature = data[:, i]
        missing = np.isnan(feature)
        valcount = (~missing).sum()
        stats[i][0] = feature[~missing].mean() if valcount >= 1 else math.nan
        stats[i][1] = feature[~missing].std()  if valcount >= 3 else math.nan
    return stats

def _normalize_data(norm, data):
    assert len(norm.shape) == 2
    assert len(data.shape) == 2
    assert norm.shape == (data.shape[1], 2)
    n = norm.shape[0]
    for i in range(n):
        feature = data[:, i]
        (mean, stdev) = norm[i]
        if not math.isnan(mean):
            feature -= mean
        if not math.isnan(stdev) and stdev > 0.0:
            feature /= stdev
        feature[np.isnan(feature)] = 0.0
    return data

def _dump_model_infos(*models, directory : str = None):
    if directory is None:
        return
    for model in models:
        txtfile = os.path.join(directory, model.name + '.txt')
        logging.info("Writing NN model summary file {!r} ...".format(txtfile))
        with open(txtfile, 'w') as ostr:
            model.summary(print_fn=(lambda s : print(s, file=ostr)))
        pdffile = os.path.join(directory, model.name + '.pdf')
        logging.info("Writing NN model plot to file {!r} ...".format(pdffile))
        try:
            _keras.utils.plot_model(model, show_shapes=True, to_file=pdffile)
        except ImportError as e:
            logging.critical(str(e))
