#! /usr/bin/python3
#! -*- coding:utf-8; mode:python; -*-

__all__ = [ 'do_properties' ]

import itertools
import logging
import math
import os

from .configuration import *
from .constants import *
from .errors import *
from .manager import *
from .utility import *

_DATA_FILE_SUFFIX = '.txt.gz'

_PROPERTY_PROGS = {
    Properties.ANGULAR     : 'angular',
    Properties.RDF_GLOBAL  : 'rdf-global',
    Properties.RDF_LOCAL   : 'rdf-local',
    Properties.EDGE_LENGTH : 'edge-length',
    Properties.PRINCOMP1ST : 'princomp',
    Properties.PRINCOMP2ND : 'princomp',
    Properties.TENSION     : 'tension',
}

_PROPERTY_FLAGS = {
    Properties.ANGULAR     : [ ],
    Properties.RDF_GLOBAL  : [ ],
    Properties.RDF_LOCAL   : [ ],
    Properties.EDGE_LENGTH : [ ],
    Properties.PRINCOMP1ST : [ '-1' ],
    Properties.PRINCOMP2ND : [ '-2' ],
    Properties.TENSION     : [ ],
}

assert all(p in _PROPERTY_PROGS for p in Properties)
assert all(p in _PROPERTY_FLAGS for p in Properties)

def do_properties(mngr : Manager, badlog : BadLog):
    for (prop, sizes) in mngr.config.desired_properties.items():
        kind = PropertyKinds.classify(prop)
        if kind is PropertyKinds.SCALED_HISTOGRAM:
            _compute_all_scaled_histograms(mngr, prop, sizes, badlog)
        elif kind is PropertyKinds.GLIDING_AVERAGE:
            _compute_all_gliding_averages(mngr, prop, sizes, badlog)
        else:
            raise ValueError(kind)

def _compute_all_scaled_histograms(mngr : Manager, prop : Properties, sizes : set, badlog):
    table = SQL_PROPERTY_TABLES[prop]
    for row in mngr.sql_select('Layouts'):
        graphid = row['graph']
        if _get_graph_size(mngr, graphid) not in sizes:
            continue
        layoutid = row['id']
        layoutfile = row['file']
        if mngr.sql_select(table, layout=layoutid):
            continue
        if badlog.get_bad(Actions.PROPERTIES, layoutid, prop):
            logging.notice("Skipping computation of property {:s} for layout {!s}".format(prop.name, layoutid))
            continue
        logging.info("Computing property {p:s} for layout {l!s} ...".format(p=prop.name, l=layoutid))
        directory = _get_directory(mngr, layoutid, prop, make=True)
        try:
            _compute_scaled_histograms(mngr, prop, layoutid, layoutfile, directory=directory)
        except RecoverableError as e:
            badlog.set_bad(Actions.PROPERTIES, layoutid, prop)
            logging.error("Cannot compute property {prop:s} for layout {lid!s} of graph {gid!s}: {!s}".format(
                e, prop=prop.name, lid=layoutid, gid=graphid))

def _compute_all_gliding_averages(mngr : Manager, prop : Properties, sizes : set, badlog):
    table = SQL_PROPERTY_TABLES[prop]
    for row in mngr.sql_select('Layouts'):
        graphid = row['graph']
        if _get_graph_size(mngr, graphid) not in sizes:
            continue
        layoutid = row['id']
        layoutfile = row['file']
        if mngr.sql_select(table, layout=layoutid):
            continue
        if badlog.get_bad(Actions.PROPERTIES, layoutid, prop):
            logging.notice("Skipping computation of property {:s} for layout {!s}".format(prop.name, layoutid))
            continue
        directory = _get_directory(mngr, layoutid, prop, make=True)
        try:
            logging.info("Computing property {:s} for layout {!s} ...".format(prop.name, layoutid))
            _compute_gliding_averages(mngr, prop, layoutid, layoutfile, directory=directory)
        except RecoverableError as e:
            badlog.set_bad(Actions.PROPERTIES, layoutid, prop)
            logging.error("Cannot compute property {prop:s} for layout {lid!s} of graph {gid!s}: {!s}".format(
                e, prop=prop.name, lid=layoutid, gid=graphid))

def _compute_scaled_histograms(mngr : Manager,prop : Properties, layoutid : Id, layoutfile : str, directory : str):
    with mngr.make_tempdir() as tempdir:
        cmd = [ _get_executable(mngr, prop) ]
        cmd.extend(_PROPERTY_FLAGS[prop])
        cmd.append('--kernel=BOXED')
        cmd.extend('--bins={:d}'.format(b) for b in FIXED_COUNT_BINS)
        cmd.append('--output={:s}'.format(os.path.join(tempdir, 'histogram-%' + _DATA_FILE_SUFFIX)))
        cmd.append('--meta=STDIO')
        cmd.append(layoutfile)
        meta = mngr.call_graphstudy_tool(cmd, meta='stdout', deterministic=True)
        _insert_scaled_histogram(mngr, prop, layoutid, meta, directory=directory, tempdir=tempdir)

def _compute_gliding_averages(mngr : Manager, prop : Properties, layoutid : Id, layoutfile : str, directory : str):
    with mngr.make_tempdir() as tempdir:
        cmd = [ _get_executable(mngr, prop) ]
        cmd.extend(_PROPERTY_FLAGS[prop])
        cmd.append('--output={:s}'.format(os.path.join(tempdir, 'gliding-%' + _DATA_FILE_SUFFIX)))
        cmd.append('--meta={:s}'.format('STDIO'))
        cmd.append('--kernel={:s}'.format('GAUSSIAN'))
        cmd.append(layoutfile)
        meta = mngr.call_graphstudy_tool(cmd, meta='stdout', deterministic=True)
        _insert_gliding_averages(mngr, prop, layoutid, meta, directory=directory, tempdir=tempdir)

def _insert_scaled_histogram(
        mngr : Manager, prop : Properties, layoutid : Id, meta : dict, directory : str, tempdir : str
):
    table = SQL_PROPERTY_TABLES[prop]
    values = {
        'layout'            : layoutid,
        'size'              : meta['size'],
        'minimum'           : meta['minimum'],
        'maximum'           : meta['maximum'],
        'mean'              : meta['mean'],
        'rms'               : meta['rms'],
        'entropy-intercept' : meta.get('entropy-intercept'),
        'entropy-slope'     : meta.get('entropy-slope'),
    }
    component = meta.get('component')
    if component is not None:
        values['component@0'] = component[0]
        values['component@1'] = component[1]
    for data in meta['data']:
        for (key, val) in data.items():
            if isinstance(val, float) and not math.isfinite(val):
                raise RecoverableError("JSON data contains non-finite value: {:s} = {!r}".format(key, val))
    renamings = dict()
    with mngr.sql_ctx as curs:
        mngr.sql_insert_curs(curs, table, **values)
        for data in meta['data']:
            filename = renamings[data['filename']] = _temporary_to_persistent_name(
                data['filename'], tempdir=tempdir, directory=directory)
            mngr.sql_insert_curs(curs, 'ScaledHistograms',
                                 property=prop,
                                 layout=layoutid,
                                 file=filename,
                                 binning=enum_from_json(Binnings, data.get('binning')),
                                 bincount=data['bincount'],
                                 binwidth=data['binwidth'],
                                 entropy=data['entropy'],
            )
    _rename_files_batch(renamings)

def _insert_gliding_averages(
        mngr : Manager, prop : Properties, layoutid : Id, meta : dict, directory : str, tempdir : str
):
    table = SQL_PROPERTY_TABLES[prop]
    renamings = dict()
    for data in meta['data']:
        for (key, val) in data.items():
            if isinstance(val, float) and not math.isfinite(val):
                raise RecoverableError("JSON data contains non-finite value: data.{:s} = {!r}".format(key, val))
        filename = renamings[data['filename']] = _temporary_to_persistent_name(
            data['filename'], tempdir=tempdir, directory=directory)
        kwargs = {
            'layout'            : layoutid,
            'file'              : filename,
            'vicinity'          : data['vicinity'],
            'size'              : data['size'],
            'minimum'           : data['minimum'],
            'maximum'           : data['maximum'],
            'mean'              : data['mean'],
            'rms'               : data['rms'],
            'sigma'             : data['sigma'],
            'points'            : data['points'],
            'entropy'           : data.get('entropy'),
            'entropy-intercept' : meta.get('entropy-intercept'),
            'entropy-slope'     : meta.get('entropy-slope'),
            'max-x'             : data['max-x'],
            'max-y'             : data['max-y'],
        }
        mngr.sql_insert(table, **kwargs)
    _rename_files_batch(renamings)

def _get_graph_size(mngr : Manager, graphid : id) -> GraphSizes:
    return GraphSizes.classify(get_one(row['nodes'] for row in mngr.sql_select('Graphs', id=graphid)))

def _get_directory(mngr : Manager, layoutid : Id, prop : Properties, make : bool = False) -> str:
    fullid = str(layoutid)
    (head, tail) = (fullid[:2], fullid[2:])
    directory = os.path.join(mngr.propsdir, head, tail, enum_to_json(prop))
    if make:
        os.makedirs(directory, exist_ok=True)
    return directory

def _get_executable(mngr : Manager, prop : Properties):
    return os.path.join(mngr.abs_execdir, 'src', 'properties', _PROPERTY_PROGS[prop])

def _temporary_to_persistent_name(tempname : str, tempdir : str, directory : str) -> str:
    relative = os.path.relpath(tempname, start=tempdir)
    return os.path.join(directory, relative)

def _rename_files_batch(renamings : dict):
    for (old, new) in renamings.items():
        logging.debug("Renaming file {!r} to {!r}".format(old, new))
        os.rename(old, new)
