-- -*- coding:utf-8; mode:sql; sql-product:sqlite; -*- --

PRAGMA foreign_keys = ON;

CREATE TABLE IF NOT EXISTS `Graphs` (
    `id`        BLOB_ID         NOT NULL UNIQUE PRIMARY KEY,
    `generator` ENUM_GENERATORS NOT NULL,
    `file`      TEXT            NOT NULL UNIQUE,
    `nodes`     INTEGER         NOT NULL CHECK (`nodes` >= 0),
    `edges`     INTEGER         NOT NULL CHECK (`edges` >= 0),
    `native`    LOGICAL         NOT NULL CHECK (`native` IN (0, 1)),
    `seed`      BLOB
);

CREATE TABLE IF NOT EXISTS `Layouts` (
    `id`     BLOB_ID        NOT NULL UNIQUE PRIMARY KEY,
    `graph`  BLOB_ID        NOT NULL REFERENCES `Graphs` ON DELETE CASCADE,
    `layout` ENUM_LAYOUTS,
    `file`   TEXT           NOT NULL UNIQUE,
    `width`  REAL           CHECK (`width`  ISNULL OR `width`  >= 0.0),
    `height` REAL           CHECK (`height` ISNULL OR `height` >= 0.0),
    `seed`   BLOB
);

CREATE INDEX IF NOT EXISTS `IndexLayoutsByGraph` ON `Layouts` (`graph`);

CREATE TABLE IF NOT EXISTS `WorseLayouts` (
    `id`     BLOB_ID         NOT NULL UNIQUE REFERENCES `Layouts` ON DELETE CASCADE,
    `parent` BLOB_ID         NOT NULL        REFERENCES `Layouts` ON DELETE CASCADE,
    `method` ENUM_LAY_WORSE  NOT NULL,
    `rate`   REAL            NOT NULL CHECK (`rate` BETWEEN 0.0 AND 1.0)
);

CREATE INDEX IF NOT EXISTS `IndexWorseLayoutsById`     ON `WorseLayouts` (`id`);
CREATE INDEX IF NOT EXISTS `IndexWorseLayoutsByParent` ON `WorseLayouts` (`parent`);

CREATE TABLE IF NOT EXISTS `InterLayouts` (
    `id`        BLOB_ID         NOT NULL UNIQUE REFERENCES `Layouts` ON DELETE CASCADE,
    `parent1st` BLOB_ID         NOT NULL        REFERENCES `Layouts` ON DELETE CASCADE,
    `parent2nd` BLOB_ID         NOT NULL        REFERENCES `Layouts` ON DELETE CASCADE,
    `method`    ENUM_LAY_INTER  NOT NULL,
    `rate`      REAL            NOT NULL CHECK (`rate` BETWEEN 0.0 AND 1.0)
);

CREATE INDEX IF NOT EXISTS `IndexInterLayoutsById`        ON `InterLayouts` (`id`);
CREATE INDEX IF NOT EXISTS `IndexInterLayoutsByParent1st` ON `InterLayouts` (`parent1st`);
CREATE INDEX IF NOT EXISTS `IndexInterLayoutsByParent2nd` ON `InterLayouts` (`parent2nd`);

CREATE TABLE IF NOT EXISTS `PrinComp1st` (
    `layout`            BLOB_ID NOT NULL UNIQUE REFERENCES `Layouts` ON DELETE CASCADE,
    `size`              INTEGER NOT NULL CHECK (`size` >= 0),
    `minimum`           REAL    NOT NULL,
    `maximum`           REAL    NOT NULL,
    `mean`              REAL    NOT NULL,
    `rms`               REAL    NOT NULL,
    `entropy-intercept` REAL    NOT NULL,
    `entropy-slope`     REAL    NOT NULL,
    `component@0`       REAL    NOT NULL,
    `component@1`       REAL    NOT NULL,
    CHECK (
        (`component@0` = 0.0 AND `component@1` = 0.0)
        OR abs(`component@0` * `component@0` + `component@1` * `component@1` - 1.0) <= 1.0E-3
    )
);

CREATE INDEX IF NOT EXISTS `IndexPrinComp1stByLayout` ON `PrinComp1st` (`layout`);

CREATE TABLE IF NOT EXISTS `PrinComp2nd` (
    `layout`            BLOB_ID NOT NULL UNIQUE REFERENCES `Layouts` ON DELETE CASCADE,
    `size`              INTEGER NOT NULL CHECK (`size` >= 0),
    `minimum`           REAL    NOT NULL,
    `maximum`           REAL    NOT NULL,
    `mean`              REAL    NOT NULL,
    `rms`               REAL    NOT NULL,
    `entropy-intercept` REAL    NOT NULL,
    `entropy-slope`     REAL    NOT NULL,
    `component@0`       REAL    NOT NULL,
    `component@1`       REAL    NOT NULL,
    CHECK (
        (`component@0` = 0.0 AND `component@1` = 0.0)
        OR abs(`component@0` * `component@0` + `component@1` * `component@1` - 1.0) <= 1.0E-3
    )
);

CREATE INDEX IF NOT EXISTS `IndexPrinComp2ndByLayout` ON `PrinComp2nd` (`layout`);

CREATE TABLE IF NOT EXISTS `Angular` (
    `layout`            BLOB_ID NOT NULL UNIQUE REFERENCES `Layouts` ON DELETE CASCADE,
    `size`              INTEGER NOT NULL CHECK (`size` >= 0),
    `minimum`           REAL    NOT NULL CHECK (`minimum` BETWEEN 0.0 AND 6.2831853072),
    `maximum`           REAL    NOT NULL CHECK (`maximum` BETWEEN 0.0 AND 6.2831853072),
    `mean`              REAL    NOT NULL CHECK (`mean`    BETWEEN 0.0 AND 6.2831853072),
    `rms`               REAL    NOT NULL CHECK (`rms`     BETWEEN 0.0 AND 6.2831853072),
    `entropy-intercept` REAL    NOT NULL,
    `entropy-slope`     REAL    NOT NULL
);

CREATE INDEX IF NOT EXISTS `IndexAngularByLayout` ON `Angular` (`layout`);

CREATE TABLE IF NOT EXISTS `EdgeLength` (
    `layout`            BLOB_ID  NOT NULL UNIQUE REFERENCES `Layouts` ON DELETE CASCADE,
    `size`              INTEGER  NOT NULL CHECK (`size`    >= 0  ),
    `minimum`           REAL     NOT NULL CHECK (`minimum` >= 0.0),
    `maximum`           REAL     NOT NULL CHECK (`maximum` >= 0.0),
    `mean`              REAL     NOT NULL CHECK (`mean`    >= 0.0),
    `rms`               REAL     NOT NULL CHECK (`rms`     >= 0.0),
    `entropy-intercept` REAL     NOT NULL,
    `entropy-slope`     REAL     NOT NULL
);

CREATE INDEX IF NOT EXISTS `IndexEdgeLengthByLayout` ON `EdgeLength` (`layout`);

CREATE TABLE IF NOT EXISTS `Tension` (
    `layout`            BLOB_ID NOT NULL UNIQUE REFERENCES `Layouts` ON DELETE CASCADE,
    `size`              INTEGER NOT NULL CHECK (`size`    >= 0),
    `minimum`           REAL    NOT NULL CHECK (`minimum` >= 0.0),
    `maximum`           REAL    NOT NULL CHECK (`maximum` >= 0.0),
    `mean`              REAL    NOT NULL CHECK (`mean`    >= 0.0),
    `rms`               REAL    NOT NULL CHECK (`rms`     >= 0.0),
    `entropy-intercept` REAL    NOT NULL,
    `entropy-slope`     REAL    NOT NULL
);

CREATE INDEX IF NOT EXISTS `IndexTensionByLayout` ON `Tension` (`layout`);

CREATE TABLE IF NOT EXISTS `GlobalRdf` (
    `layout`            BLOB_ID NOT NULL UNIQUE REFERENCES `Layouts` ON DELETE CASCADE,
    `size`              INTEGER NOT NULL CHECK (`size`    >= 0  ),
    `minimum`           REAL    NOT NULL CHECK (`minimum` >= 0.0),
    `maximum`           REAL    NOT NULL CHECK (`maximum` >= 0.0),
    `mean`              REAL    NOT NULL CHECK (`mean`    >= 0.0),
    `rms`               REAL    NOT NULL CHECK (`rms`     >= 0.0),
    `entropy-intercept` REAL    NOT NULL,
    `entropy-slope`     REAL    NOT NULL
);

CREATE INDEX IF NOT EXISTS `IndexGlobalRdfByLayout` ON `GlobalRdf` (`layout`);

-- TODO: Refactor this unholy mess.
CREATE TABLE IF NOT EXISTS `LocalRdf` (
    `layout`            BLOB_ID NOT NULL REFERENCES `Layouts` ON DELETE CASCADE,
    `file`              TEXT    NOT NULL UNIQUE,
    `vicinity`          INTEGER NOT NULL CHECK (`vicinity` >  0  ),
    `size`              INTEGER NOT NULL CHECK (`size`     >= 0  ),
    `minimum`           REAL    NOT NULL CHECK (`minimum`  >= 0.0),
    `maximum`           REAL    NOT NULL CHECK (`maximum`  >= 0.0),
    `mean`              REAL    NOT NULL CHECK (`mean`     >= 0.0),
    `rms`               REAL    NOT NULL CHECK (`rms`      >= 0.0),
    `sigma`             REAL    NOT NULL CHECK (`sigma`    >  0.0),
    `points`            INTEGER NOT NULL CHECK (`points`   >  0  ),
    `max-x`             REAL    NOT NULL CHECK (`max-x`    >= 0.0),
    `max-y`             REAL    NOT NULL CHECK (`max-y`    >= 0.0),
    `entropy`           REAL,   -- Nothing can be checked as long as we have no real entropy.
    `entropy-intercept` REAL,
    `entropy-slope`     REAL,
    UNIQUE (`layout`, `vicinity`)
);

CREATE INDEX IF NOT EXISTS `IndexLocalRdfByLayout` ON `LocalRdf` (`layout`);

CREATE TABLE IF NOT EXISTS `ScaledHistograms` (
    `property` ENUM_PROPERTIES NOT NULL,
    `layout`   BLOB_ID         NOT NULL REFERENCES `Layouts` ON DELETE CASCADE,
    `file`     TEXT            NOT NULL UNIQUE,
    `binning`  ENUM_BINNINGS   NOT NULL,
    `bincount` INTEGER         NOT NULL CHECK (`bincount` >  0  ),
    `binwidth` REAL            NOT NULL CHECK (`binwidth` >  0.0),
    `entropy`  REAL            NOT NULL CHECK (`entropy`  >= 0.0),
    UNIQUE (`property`, `layout`, `binning`, `bincount`)
);

CREATE INDEX IF NOT EXISTS `IndexScaledHistogramsByLayout` ON `ScaledHistograms` (`layout`);

CREATE TABLE IF NOT EXISTS `Metrics` (
    `metric`   ENUM_METRICS NOT NULL,
    `layout`   BLOB_ID      NOT NULL REFERENCES `Layouts` ON DELETE CASCADE,
    `value`    REAL         NOT NULL,
    UNIQUE (`metric`, `layout`)
);

CREATE TABLE IF NOT EXISTS `NNTestScore` (
    `lhs`      BLOB_ID  REFERENCES `Layouts` ON DELETE SET NULL,
    `rhs`      BLOB_ID  REFERENCES `Layouts` ON DELETE SET NULL,
    `expected` REAL     NOT NULL CHECK (`expected` BETWEEN -1.0 AND +1.0),
    `actual`   REAL     NOT NULL CHECK (`actual` BETWEEN -1.0 AND +1.0)
);

CREATE INDEX IF NOT EXISTS `IndexNNTestScoreByLhs` ON `NNTestScore` (`lhs`);
CREATE INDEX IF NOT EXISTS `IndexNNTestScoreByRhs` ON `NNTestScore` (`rhs`);

CREATE TABLE IF NOT EXISTS `ToolPerformance` (
    `tool` TEXT NOT NULL,
    `time` REAL NOT NULL CHECK (`time` >= 0.0)
);
