#! /usr/bin/python3
#! -*- coding:utf-8; mode:python; -*-

__all__ = [ 'get_sql_view_code' ]

import argparse
import shlex
import sys

from .constants import *

_COLUMNS_SCALED_HISTOGRAM = [ 'mean', 'rms', 'entropy-intercept', 'entropy-slope' ]
_COLUMNS_GLIDING_AVERAGE  = [ 'mean', 'rms', 'entropy' ]

def get_sql_view_code(properties, force : bool = False):
    scriptlines = list()
    scriptlines.append("-- -*- coding:utf-8; mode:sql; sql-product:sqlite; -*- --")
    scriptlines.append("")
    scriptlines.append("-- This is a generated SQLite script produced by the following command:")
    scriptlines.append("-- " + ' '.join(map(shlex.quote, _get_cmdline(force))))
    scriptlines.append("")
    if force:
        scriptlines.append("DROP VIEW IF EXISTS " + _quote('GraphFeatures') + ";")
        scriptlines.append("")
    scriptlines.append("CREATE VIEW IF NOT EXISTS " + _quote('GraphFeatures') + " AS SELECT")
    _append_select_item(scriptlines, 'id',    dst='graph')
    _append_select_item(scriptlines, 'nodes', dst='pn', func='logarithm')
    _append_select_item(scriptlines, 'edges', dst='pm', func='logarithm')
    scriptlines[-1] = scriptlines[-1].rstrip(',')
    scriptlines.append("FROM " + _quote('Graphs') + ";")
    scriptlines.append("")
    if force:
        scriptlines.append("DROP VIEW IF EXISTS " + _quote('LayoutFeatures') + ";")
        scriptlines.append("")
    scriptlines.append("CREATE VIEW IF NOT EXISTS " + _quote('LayoutFeatures') + " AS SELECT")
    _append_select_item(scriptlines, ('Layouts', 'id'), dst='layout')
    if Properties.PRINCOMP1ST in properties:
        _append_select_item(scriptlines, ('PrinComp1st', 'component@0'), dst='major@0')
        _append_select_item(scriptlines, ('PrinComp1st', 'component@1'), dst='major@1')
    if Properties.PRINCOMP2ND in properties:
        _append_select_item(scriptlines, ('PrinComp2nd', 'component@0'), dst='minor@0')
        _append_select_item(scriptlines, ('PrinComp2nd', 'component@1'), dst='minor@1')
    for prop in sorted(PROPERTIES_SCALED_HISTOGRAM & properties):
        table = SQL_PROPERTY_TABLES[prop]
        scriptlines.append("\t-- {:s}".format(prop.name))
        for column in _COLUMNS_SCALED_HISTOGRAM:
            _append_select_item(scriptlines, (table, column), dst='{:s}::{:s}'.format(table, column))
    for prop in sorted(PROPERTIES_GLIDING_AVERAGE & properties):
        table = SQL_PROPERTY_TABLES[prop]
        for i in VICINITIES_LOG_2:
            alias = '{:s}{:d}'.format(table, i)
            scriptlines.append("\t-- {:s} (vicinity = {:d})".format(prop.name, 2**i))
            for column in _COLUMNS_GLIDING_AVERAGE:
                _append_select_item(scriptlines, (alias, column), dst='{:s}::{:s}'.format(alias, column))
    scriptlines[-1] = scriptlines[-1].rstrip(',')
    scriptlines.append("FROM " + _quote('Layouts'))
    for prop in sorted(PROPERTIES_SCALED_HISTOGRAM & properties):
        table = SQL_PROPERTY_TABLES[prop]
        lhs = _quote('Layouts', 'id')
        rhs = _quote(table, 'layout')
        scriptlines.append("LEFT JOIN {:24s} ON {:s} = {:s}".format(_quote(table), lhs, rhs))
    for prop in sorted(PROPERTIES_GLIDING_AVERAGE & properties):
        table = SQL_PROPERTY_TABLES[prop]
        for i in VICINITIES_LOG_2:
            alias = '{:s}{:d}'.format(table, i)
            lhs = _quote('Layouts', 'id')
            rhs = _quote(alias, 'layout')
            scriptlines.append("LEFT JOIN (SELECT * FROM {:s} WHERE {:s} = {:6d}) AS {:s} ON {:s} = {:s}".format(
                _quote(table), _quote('vicinity'), 2**i, _quote(alias), lhs, rhs))
    scriptlines.append(";")
    scriptlines.append("")
    return '\n'.join(scriptlines)

def _append_select_item(scriptlines, src, dst=None, func=None):
    source = _quote(src) if isinstance(src, str) else _quote(*src)
    if func is not None:
        source = '{:s}({:s})'.format(func, source)
    if dst is not None:
        scriptlines.append("\t{:36s} AS {:s},".format(source, _quote(dst)))
    else:
        scriptlines.append("\t{:36s},".format(source))

def _quote(*names):
    assert len(names) > 0
    assert all('`' not in name for name in names)
    return '.'.join('`{:s}`'.format(name) for name in names)

def _get_cmdline(force):
    if __name__ == '__main__':
        return sys.argv
    else:
        cmd = [ sys.executable, '-m', __name__ ]
        if force:
            cmd.append('--force')
        return cmd

if __name__ == '__main__':
    ap = argparse.ArgumentParser(description="Prints SQLite view creation code for debugging.")
    ap.add_argument('--force', '-f', action='store_true', help="drop any existing views")
    ns = ap.parse_args()
    print(get_sql_view_code(frozenset(Properties), force=ns.force))
