#! /usr/bin/python3
#! -*- coding:utf-8; mode:python; -*-

__all__ = [
    'BZ2_FILTER',
    'GNUPLOT',
    'GZIP_FILTER',
    'IMAGE_MAGICK',
    'find_all_tools_eagerly',
    'find_tool_lazily',
    'get_cache_directory',
]

import logging
import os
import shlex
import shutil

from .errors import *
from .utility import *

GNUPLOT      = object()
IMAGE_MAGICK = object()
GZIP_FILTER  = object()
BZ2_FILTER   = object()

_TOOL_CACHE = dict()

_TOOL_SPECS = {
    GNUPLOT      : ('GNUPLOT',      'gnuplot'  ),
    IMAGE_MAGICK : ('IMAGE_MAGICK', 'convert'  ),
    GZIP_FILTER  : ('ZCAT',         'gzip -dc' ),
    BZ2_FILTER   : ('BZCAT',        'bzip2 -dc'),
}

def find_tool_lazily(tool):
    try:
        return _TOOL_CACHE[tool]
    except KeyError:
        pass
    (envvar, default) = _TOOL_SPECS[tool]
    result = _find_tool(envvar, default)
    _TOOL_CACHE[tool] = result
    return result

def find_all_tools_eagerly():
    _get_cache_directory()
    for (key, val) in _TOOL_SPECS.items():
        _TOOL_CACHE[key] = _find_tool(*val)

def _find_tool(envvar, default):
    for prefix in [ 'MSC_', '' ]:
        envval = os.getenv(prefix + envvar)
        if envval is not None:
            tokens = envval.split()
            if not tokens or not os.path.isabs(tokens[0]):
                raise SanityError("Environment variable {:s} not set to an absolute path".format(envvar))
            logging.info("Using {:s}={!r} (environment)".format(envvar, envval))
            return envval
    (prog, *flags) = default.split()
    prog = shutil.which(prog)
    if not prog:
        logging.warning("Cannot find tool {:s} (falling back to {!r})".format(envvar, default))
        return default
    elif shlex.quote(prog) != prog:
        raise SanityError("Cowardly refusing to use unsafe external tool path: {!r}".format(prog))
    else:
        cmdstr = ' '.join([prog, *flags])
        logging.info("Using {:s}={!r} (default)".format(envvar, cmdstr))
        return cmdstr

_CACHE_DIRECTORY = list()  # poor-man's optional type that can hold None as a value

def get_cache_directory(fallback : str = None):
    if not _CACHE_DIRECTORY:
        _CACHE_DIRECTORY.append(_get_cache_directory())
    directory = get_one(_CACHE_DIRECTORY)
    return directory if directory is not None else fallback

def _get_cache_directory():
    envvar = 'MSC_CACHE_DIR'
    directory = os.getenv(envvar)
    if not directory:
        return None
    if not os.path.isabs(directory) or not os.path.isdir(directory):
        raise ConfigError("Not an absolute path of an existing directory: {!s}={!r}".format(envvar, directory))
    if not os.path.exists(os.path.join(directory, 'CACHEDIR.TAG')):
        logging.warning("Cache directory {!r} contains no {:s} file (see {:s} for an explanation)".format(
            directory, 'CACHEDIR.TAG', 'http://www.brynosaurus.com/cachedir/'))
    logging.info("Using cache directory {:s}={!r}".format(envvar, directory))
    return directory
