#! /usr/bin/python3
#! -*- coding:utf-8; mode:python; -*-

__all__ = [
    'append_common_all_constants',
    'append_common_graph',
    'append_common_property',
    'get_bool_from_query',
    'get_informal_layout_name',
    'get_informal_layout_name_curs',
]

from . import *
from ..constants import *
from ..utility import *

_ENUM_TAG_NAMES = {
    Binnings   : ('all-binnings',   'binning'),
    Generators : ('all-generators', 'generator'),
    GraphSizes : ('all-sizes',      'size'),
    Layouts    : ('all-layouts',    'layout'),
    LayInter   : ('all-lay-inter',  'lay-inter'),
    LayWorse   : ('all-lay-worse',  'lay-worse'),
    Properties : ('all-properties', 'property'),
    TestStatus : ('all-status',     'status'),
}

def append_common_all_constants(parent, *enumtypes):
    for enumtype in enumtypes:
        (ptag, ctag) = _ENUM_TAG_NAMES[enumtype]
        with Child(parent, ptag) as node:
            for const in enumtype:
                append_child(node, ctag, key=str(const.value)).text = const.name

def append_common_graph(parent, row, tag='graph'):
    gen = row['generator']
    size = GraphSizes.classify(row['nodes'])
    node = append_child(parent, tag, id=str(row['id'])) if tag is not None else parent
    append_child(node, 'generator').text = gen.name
    append_child(node, 'size').text = size.name
    append_child(node, 'nodes').text = fmtnum(row['nodes'])
    append_child(node, 'edges').text = fmtnum(row['edges'])

def append_common_property(prop, parent, rows, tagname='property-data'):
    kind = PropertyKinds.classify(prop)
    commonkeys = { 'size', 'minimum', 'maximum', 'mean', 'rms' }
    if not rows:
        return
    elif kind is PropertyKinds.SCALED_HISTOGRAM:
        with Child(parent, tagname, property=prop.name, kind=kind.name) as child:
            for row in rows:
                with Child(child, 'histogram') as histo:
                    binning = row['binning']
                    append_child(histo, 'binning').text = binning.name
                    for key in commonkeys | { 'bincount', 'binwidth' }:
                        append_child(histo, key).text = fmtnum(row[key])
    elif kind is PropertyKinds.GLIDING_AVERAGE:
        with Child(parent, tagname, property=prop.name, kind=kind.name) as child:
            for row in rows:
                with Child(child, 'density') as densi:
                    for key in commonkeys | { 'vicinity', 'sigma', 'points' }:
                        append_child(densi, key).text = fmtnum(row[key])
    else:
        raise AssertionError(prop)

def get_bool_from_query(query, key, default=False):
    try:
        [ value ] = query[key]
    except KeyError:
        return default
    except ValueError:
        raise HttpError(http.HTTPStatus.BAD_REQUEST, "Parameter {!r} must be passed at most once.".format(key))
    try:
        return parse_bool(value)
    except ValueError:
        raise HttpError(http.HTTPStatus.BAD_REQUEST, "Parameter {!r} must be set to a boolean.".format(key))

def get_informal_layout_name(mngr, layoutid) -> str:
    with mngr.sql_ctx as curs:
        return get_informal_layout_name_curs(mngr, curs, layoutid)

def get_informal_layout_name_curs(mngr, curs, layoutid) -> str:
    assert isinstance(layoutid, Id)
    proper = get_one(mngr.sql_select_curs(curs, 'Layouts', id=layoutid))['layout']
    if proper is not None:
        return proper.name
    for row in mngr.sql_select_curs(curs, 'InterLayouts', id=layoutid):
        name1 = get_informal_layout_name_curs(mngr, curs, row['parent1st'])
        name2 = get_informal_layout_name_curs(mngr, curs, row['parent2nd'])
        return "({:s}){:d}:{:s}({:s})".format(name1, round(100.0 * row['rate']), row['method'].name, name2)
    for row in mngr.sql_select_curs(curs, 'WorseLayouts', id=layoutid):
        name = get_informal_layout_name_curs(mngr, curs, row['parent'])
        return "({:s})+{:d}:{:s}".format(name, round(100.0 * row['rate']), row['method'].name)
    return "???"
