#! /usr/bin/python3
#! -*- coding:utf-8; mode:python; -*-

__all__ = [ 'serve' ]

import collections
import datetime
import http
import itertools
import logging
import math
import random
import statistics
import time
import typing
import urllib.parse

from . import *
from ..constants import *
from ..model import *
from ..nn import *
from ..quarry import *
from ..utility import *
from .impl_common import *

def _serve_features(this, url=None):
    query = urllib.parse.parse_qs(url.query)
    (lhs, rhs) = _get_layout_ids_from_query(query, manager=this.server.graphstudy_manager)
    try:
        with open(this.server.graphstudy_manager.nn_features, 'rb') as istr:
            timestamp, features = unpickle_objects(istr, typing.List[NNTestResult], error=_500)
    except FileNotFoundError:
        raise HttpError(http.HTTPStatus.NOT_FOUND, "Apparently, no model was built yet.")
    if lhs and rhs:
        (graphfeatrow, lhsfeatrow, rhsfeatrow) = _load_feature_rows(this.server.graphstudy_manager, lhs, rhs)
    root = ET.Element('root')
    append_child(root, 'timestamp').text = rfc5322(timestamp)
    if lhs and rhs:
        with Child(root, 'this') as node:
            append_child(node, 'graph').text = str(graphfeatrow[0])
            append_child(node, 'lhs').text = str(lhsfeatrow[0])
            append_child(node, 'rhs').text = str(rhsfeatrow[0])
    for (of, vector) in features.items():
        with Child(root, 'features', of=of) as node:
            for (i, feat) in enumerate(vector):
                with Child(node, 'feature', index=fmtnum(i)) as child:
                    append_child(child, 'name').text = feat.name
                    append_child(child, 'mean').text = fmtnum(feat.mean)
                    append_child(child, 'stdev').text = fmtnum(feat.stdev)
                    if lhs and rhs:
                        def getit(row):
                            diff = value_or(row[i + 1], math.nan) - feat.mean
                            if feat.stdev > 0.0: return diff / feat.stdev
                            return math.nan if diff == 0.0 else diff * math.inf
                        if of == 'graph':
                            append_child(child, 'this').text = fmtnum(getit(graphfeatrow))
                        if of == 'layout':
                            (lhsfeat, rhsfeat) = (getit(lhsfeatrow), getit(rhsfeatrow))
                            append_child(child, 'this-lhs').text = fmtnum(lhsfeat)
                            append_child(child, 'this-rhs').text = fmtnum(rhsfeat)
                            append_child(child, 'this-diff').text = fmtnum(lhsfeat - rhsfeat)
    this.send_tree_xml(ET.ElementTree(root), transform='/xslt/nn-features.xsl')

def _load_feature_rows(mngr, lhs, rhs):
    assert isinstance(lhs, Id) and isinstance(rhs, Id)
    with mngr.sql_ctx as curs:
        gridl = get_one_or(r['graph'] for r in mngr.sql_select_curs(curs, 'Layouts', id=lhs))
        gridr = get_one_or(r['graph'] for r in mngr.sql_select_curs(curs, 'Layouts', id=rhs))
        if gridl is None:
            raise HttpError(http.HTTPStatus.NOT_FOUND, "No layout with ID {!s} found.".format(lhs))
        if gridr is None:
            raise HttpError(http.HTTPStatus.NOT_FOUND, "No layout with ID {!s} found.".format(rhs))
        if gridl != gridr:
            raise HttpError(
                http.HTTPStatus.BAD_REQUEST,
                "The layouts {!s} and {!s} are from different graphs ({!s} and {!s}).".format(lhs, rhs, gridl, gridr)
            )
        graphrow = get_one(mngr.sql_select_curs(curs, 'GraphFeatures', graph=gridr))
        lhsrow = get_one(mngr.sql_select_curs(curs, 'LayoutFeatures', layout=lhs))
        rhsrow = get_one(mngr.sql_select_curs(curs, 'LayoutFeatures', layout=rhs))
        return (graphrow, lhsrow, rhsrow)

def _serve_testscore(this, url=None):
    query = urllib.parse.parse_qs(url.query)
    graphical = get_bool_from_query(query, 'graphical')
    results = [
        NNTestResult(r['lhs'], r['rhs'], r['expected'], r['actual'])
        for r in this.server.graphstudy_manager.sql_select('NNTestScore')
    ]
    if not results:
        raise HttpError(http.HTTPStatus.NOT_FOUND, "Apparently, no model was built yet.")
    condmean  = math.nan if len(results) < 1 else statistics.mean( (tr.expected for tr in results))
    condstdev = math.nan if len(results) < 3 else statistics.stdev((tr.expected for tr in results), xbar=condmean)
    predmean  = math.nan if len(results) < 1 else statistics.mean( (tr.actual   for tr in results))
    predstdev = math.nan if len(results) < 3 else statistics.stdev((tr.actual   for tr in results), xbar=predmean)
    errmean   = math.nan if len(results) < 1 else statistics.mean( (tr.error    for tr in results))
    errstdev  = math.nan if len(results) < 3 else statistics.stdev((tr.error    for tr in results), xbar=errmean)
    laymap = dict()
    for tr in results:
        laymap[tr.lhs] = None
        laymap[tr.rhs] = None
    for layoutid in laymap.keys():
        laymap[layoutid] = get_informal_layout_name(this.server.graphstudy_manager, layoutid)
    root = ET.Element('root')
    append_common_all_constants(root, TestStatus)
    with Child(root, 'summary') as summary:
        with Child(summary, 'condition-bias') as node:
            append_child(node, 'mean').text = fmtnum(condmean)
            append_child(node, 'stdev').text = fmtnum(condstdev)
        with Child(summary, 'prediction-bias') as node:
            append_child(node, 'mean').text = fmtnum(predmean)
            append_child(node, 'stdev').text = fmtnum(predstdev)
        with Child(summary, 'error-bias') as node:
            append_child(node, 'mean').text = fmtnum(errmean)
            append_child(node, 'stdev').text = fmtnum(errstdev)
    with Child(root, 'test-cases') as node:
        for tr in results:
            with Child(node, 'test-case', status=tr.status.name) as child:
                with Child(child, 'lhs') as lhs:
                    if laymap[tr.lhs] is not None: lhs.set('informal', laymap[tr.lhs])
                    lhs.text = str(tr.lhs)
                with Child(child, 'rhs') as rhs:
                    if laymap[tr.rhs] is not None: rhs.set('informal', laymap[tr.rhs])
                    rhs.text = str(tr.rhs)
                append_child(child, 'expected').text = fmtnum(tr.expected)
                append_child(child, 'actual').text = fmtnum(tr.actual)
                append_child(child, 'error').text = fmtnum(tr.error)
    stylesheet = '/xslt/nn-testscore-graphical.xsl' if graphical else '/xslt/nn-testscore.xsl'
    this.send_tree_xml(ET.ElementTree(root), transform=stylesheet)

def _serve_demo(this, url):
    query = urllib.parse.parse_qs(url.query)
    pairs = list()
    explicit = _get_layout_ids_from_query(query, manager=this.server.graphstudy_manager)
    randcount = _get_random_count_from_query(query)
    if explicit != (None, None):
        pairs.append(explicit)
    if randcount > 0:
        rand = random.Random()
        pairs.extend(_get_random_layout_pairs(this.server.graphstudy_manager, rand, randcount))
        if not pairs:
            raise HttpError(http.HTTPStatus.NOT_FOUND, "There are no layouts to begin with.")
    root = ET.Element('root')
    if pairs:
        t0 = time.time()
        oracle = Oracle(this.server.graphstudy_manager, exception=_500)
        t1 = time.time()
        wisdom = oracle(pairs, exception=_500, bidirectional=True)
        t2 = time.time()
        with Child(root, 'timings') as node:
            append_child(node, 'load').text = fmtnum(t1 - t0)
            append_child(node, 'eval').text = fmtnum(t2 - t1)
            append_child(node, 'total').text = fmtnum(t2 - t0)
        with Child(root, 'results') as node:
            for ((lhs, rhs), (ans, san)) in zip(pairs, wisdom):
                with Child(node, 'result', random=format_bool((lhs, rhs) != explicit)) as child:
                    with Child(child, 'lhs') as lhsnode:
                        lhsnode.set('informal', get_informal_layout_name(this.server.graphstudy_manager, lhs))
                        lhsnode.text = str(lhs)
                    with Child(child, 'rhs') as rhsnode:
                        rhsnode.set('informal', get_informal_layout_name(this.server.graphstudy_manager, rhs))
                        rhsnode.text = str(rhs)
                    append_child(child, 'prediction').text = fmtnum(ans)
                    append_child(child, 'reverse').text = fmtnum(san)
    this.send_tree_xml(ET.ElementTree(root), transform='/xslt/nn-demo.xsl')

def _get_random_layout_pairs(mngr, rand, count=1):
    pairs = set()
    with mngr.sql_ctx as curs:
        graphs = [ r['id'] for r in mngr.sql_select_curs(curs, 'Graphs') ]
        while graphs and len(pairs) < count:
            graphid = rand.choice(graphs)
            layouts = sorted((r['id'] for r in mngr.sql_select_curs(curs, 'Layouts', graph=graphid)), key=Id.getkey)
            for pair in filter(lambda p : p not in pairs, itertools.combinations(layouts, 2)):
                pairs.add(pair)
                break
            else:
                graphs.remove(graphid)
    return pairs

def _get_random_count_from_query(query):
    try:
        values = list(map(int, query['random']))
        if all(n >= 0 for n in values):
            return sum(values)
    except KeyError:
        return 0
    except ValueError:
        pass
    raise HttpError(http.HTTPStatus.BAD_REQUEST, "Parameter 'random' must be set to a non-negative integer.")

def _get_layout_ids_from_query(query, manager=None):
    lhs = query.get('lhs', [ ])
    rhs = query.get('rhs', [ ])
    if not lhs and not rhs:
        return (None, None)
    elif len(lhs) > 1:
        raise HttpError(http.HTTPStatus.BAD_REQUEST, "Parameters 'lhs' must not be used more than once.")
    elif len(rhs) > 1:
        raise HttpError(http.HTTPStatus.BAD_REQUEST, "Parameters 'rhs' must not be used more than once.")
    elif len(lhs) != len(rhs):
        raise HttpError(http.HTTPStatus.BAD_REQUEST, "Parameters 'lhs' and 'rhs' must be used together.")
    elif manager is None:
        getter = lambda s : validate_layout_id(get_one(s))
        return (getter(lhs), getter(rhs))
    else:
        with manager.sql_ctx as curs:
            getter = lambda s : manager.idmatch_curs(curs, 'Layouts', get_one(s), exception=_404)
            return (getter(lhs), getter(rhs))

def _500(msg):
    return HttpError(http.HTTPStatus.INTERNAL_SERVER_ERROR, msg)

def _404(msg):
    return HttpError(http.HTTPStatus.NOT_FOUND, msg)

class _NotFound(object):

    def __call__(self, *args, **kwargs):
        raise HttpError(http.HTTPStatus.NOT_FOUND)

serve = collections.defaultdict(_NotFound)
serve['features'] = _serve_features
serve['testscore'] = _serve_testscore
serve['demo'] = _serve_demo
