#! /usr/bin/python3
#! -*- coding:utf-8; mode:python; -*-

__all__ = [ 'serve' ]

import http
import logging
import urllib.parse

from . import *
from ..constants import *
from ..utility import *
from .impl_common import *

def serve(this, graphid : Id, prop : Properties, url):
    query = urllib.parse.parse_qs(url.query)
    try:
        selected_layouts = frozenset(validate_layout_id(token) for token in query['layouts'])
    except KeyError:
        selected_layouts = None
    with this.server.graphstudy_manager.sql_ctx as curs:
        graphrow = get_one_or(this.server.graphstudy_manager.sql_select_curs(curs, 'Graphs', id=graphid))
        if not graphrow:
            raise HttpError(http.HTTPStatus.NOT_FOUND, "No graph with ID {!s} exists.".format(graphid))
        root = ET.Element('root')
        append_child(root, 'property').text = prop.name
        append_common_graph(root, graphrow)
        if prop in PROPERTIES_SCALED_HISTOGRAM:
            append_common_all_constants(root, Layouts, Properties, Binnings)
            _append_scaled_histogram_stuff(this.server.graphstudy_manager, graphid, selected_layouts, prop, root, curs)
        elif prop in PROPERTIES_GLIDING_AVERAGE:
            append_common_all_constants(root, Layouts, Properties)
            _append_gliding_average_stuff(this.server.graphstudy_manager, graphid, selected_layouts, prop, root, curs)
        else:
            raise AssertionError(p)
    this.send_tree_xml(ET.ElementTree(root), transform='/xslt/property.xsl')

def _append_scaled_histogram_stuff(mngr, graphid, selected_layouts, prop, root, curs):
    get_one(root.findall('property')).set('kind', 'SCALED_HISTOGRAM')
    with Child(root, 'all-bincounts') as abc:
        for bc in FIXED_COUNT_BINS:
            append_child(abc, 'bincount').text = fmtnum(bc)
    with Child(root, 'layouts') as los:
        all_layout_ids = set()
        for layoutrow in _unordered(mngr.sql_select_curs(curs, 'Layouts', graph=graphid, layout=object)):
            layoutid = layoutrow['id']
            all_layout_ids.add(layoutid)
            with Child(los, 'layout-data', id=str(layoutid), layout=layoutrow['layout'].name) as lo:
                kwargs = { 'property' : prop, 'layout' : layoutid, 'binning' : Binnings.SCOTT_NORMAL_REFERENCE }
                for row in mngr.sql_select_curs(curs, 'ScaledHistograms', **kwargs):
                    attrs = { 'bincount' : fmtnum(row['bincount']), 'binning' : Binnings.SCOTT_NORMAL_REFERENCE.name }
                    append_child(lo, 'entropy', **attrs).text = fmtnum(row['entropy'])
                kwargs = { 'property' : prop, 'layout' : layoutid, 'binning' : Binnings.FIXED_COUNT }
                entropy_fixed = {
                    r['bincount'] : r['entropy'] for r in
                    mngr.sql_select_curs(curs, 'ScaledHistograms', **kwargs)
                }
                for bc in FIXED_COUNT_BINS:
                    attrs = { 'bincount' : fmtnum(bc), 'binning' : Binnings.FIXED_COUNT.name }
                    try:
                        value = entropy_fixed[bc]
                    except KeyError:
                        pass
                    else:
                        append_child(lo, 'entropy', **attrs).text = fmtnum(value)
                selected = count(lo.findall('entropy')) if selected_layouts is None else layoutid in selected_layouts
                lo.set('selected', format_bool(selected))
    if selected_layouts is not None:
        for lid in selected_layouts - all_layout_ids:
            raise HttpError(http.HTTPStatus.NOT_FOUND, "No layout with ID {!s} exists.".format(graphid))

def _append_gliding_average_stuff(mngr, graphid, selected_layouts, prop, root, curs):
    get_one(root.findall('property')).set('kind', 'GLIDING_AVERAGE')
    table = SQL_PROPERTY_TABLES[prop]
    with Child(root, 'layouts') as los:
        all_layout_ids = set()
        all_vicinities = set()
        for layoutrow in _unordered(mngr.sql_select_curs(curs, 'Layouts', graph=graphid, layout=object)):
            layoutid = layoutrow['id']
            all_layout_ids.add(layoutid)
            with Child(los, 'layout-data', id=str(layoutid), layout=layoutrow['layout'].name) as lo:
                rows = mngr.sql_select_curs(curs, table, layout=layoutid)
                all_vicinities.update(r['vicinity'] for r in rows)
                for row in sorted(rows, key=(lambda r : r['vicinity'])):
                    attrs = { key : fmtnum(row[key]) for key in [ 'vicinity', 'sigma', 'points' ] }
                    append_child(lo, 'entropy', **attrs).text = fmtnum(row['entropy'])
                selected = count(lo.findall('entropy')) if selected_layouts is None else layoutid in selected_layouts
                lo.set('selected', format_bool(selected))
    with Child(root, 'all-vicinities') as avs:
        for vcn in sorted(all_vicinities):
            append_child(avs, 'vicinity').text = fmtnum(vcn)
    if selected_layouts is not None:
        for lid in selected_layouts - all_layout_ids:
            raise HttpError(http.HTTPStatus.NOT_FOUND, "No layout with ID {!s} exists.".format(graphid))

def _unordered(seq):
    """
    This function does nothing but is a hint that the iteration order is unspecified.
    """
    return seq
