#! /usr/bin/python3
#! -*- coding:utf-8; mode:python; -*-

__all__ = [ 'serve' ]

import http
import logging
import math
import os.path
import statistics
import textwrap
import urllib.parse

from . import *
from ..constants import *
from ..errors import *
from ..tools import *
from ..utility import *
from .impl_common import *

class _GN(float):

    def __init__(self, value=None):
        self.__value = math.nan if value is None else float(value)

    @property
    def value(self):
        return self.__value

    def __format__(self, fmtspec=None):
        textual = 'NaN' if not math.isfinite(self.__value) else format(self.__value, '.10E')
        if not fmtspec or fmtspec == 'G':
            return textual
        elif fmtspec.endswith('G'):
            width = int(fmtspec[:-1])
            return ''.join(' ' for i in range(max(0, width - len(textual)))) + textual
        else:
            raise ValueError(fmtspec)

def _serve_histogram(this, prop : Properties, layoutid : Id, url, format : str = None):
    if PropertyKinds.classify(prop) is not PropertyKinds.SCALED_HISTOGRAM:
        raise HttpError(http.HTTPStatus.NOT_FOUND, "No histogram for property {:s} available.".format(prop.name))
    (terminal, mimetype, charset) = _check_graphics_format(this, format)
    query = urllib.parse.parse_qs(url.query)
    serve_source = get_bool_from_query(query, 'source')
    bincount = _get_bincount_from_query(query)
    if bincount is None:
        snr = Binnings.SCOTT_NORMAL_REFERENCE
        row = get_first_or(this.server.graphstudy_manager.sql_select(
            'ScaledHistograms', property=prop, layout=layoutid, binning=snr)
        )
        if not row:
            msg = "No histogram for property {:s} with automatic binning available.".format(prop.name)
            raise HttpError(http.HTTPStatus.NOT_FOUND, msg)
    else:
        fc = Binnings.FIXED_COUNT
        row = get_first_or(this.server.graphstudy_manager.sql_select(
            'ScaledHistograms', property=prop, layout=layoutid, binning=fc, bincount=bincount)
        )
        if not row:
            msg = "No histogram for property {:s} of layout {!s} with {:d} bins ({:s}) available.".format(
                prop.name, layoutid, bincount, fc.name)
            raise HttpError(http.HTTPStatus.NOT_FOUND, msg)
    assert row is not None
    xlabel = "{0[1]:s} / {0[0]:s}".format(PROPERTY_QUANTITIES[prop])
    ylabel = "relative frequency"
    title = "{bc:,d} bins (width = {bw:,.3g} {xunit:s})".format(
        bc=row['bincount'], bw=row['binwidth'], xunit=PROPERTY_QUANTITIES[prop][0]
    )
    color = '#{:06X}'.format(PRIMARY_COLOR)
    gnuplot = find_tool_lazily(GNUPLOT)
    script = textwrap.dedent('''\
    #! {gp:s}
    #! -*- coding:utf-8; mode:gnuplot; -*-"

    set terminal {fmt:s} size 800,600
    set output
    set xrange [0 : *]
    set yrange [0 : *]
    set xlabel {xlabel!r}
    set ylabel {ylabel!r}
    set format y '%.2E'
    set style fill solid 1.0
    plot {df!r} with boxes linecolor rgb {lc!r} title {title!r}
    ''').format(gp=gnuplot, fmt=terminal, df=_gpltin(row['file']), xlabel=xlabel, ylabel=ylabel, title=title, lc=color)
    _serve_plot(this, script, mimetype, charset, source=serve_source)

def _serve_gliding_average(this, prop : Properties, layoutid : Id, url, format : str = None):
    if PropertyKinds.classify(prop) is not PropertyKinds.GLIDING_AVERAGE:
        msg = "No gliding average for property {:s} available.".format(prop.name)
        raise HttpError(http.HTTPStatus.NOT_FOUND, msg)
    assert prop is Properties.RDF_LOCAL
    (terminal, mimetype, charset) = _check_graphics_format(this, format)
    query = urllib.parse.parse_qs(url.query)
    serve_source = get_bool_from_query(query, 'source')
    vicinities = _get_vicinities_from_query(query)
    table = SQL_PROPERTY_TABLES[prop]
    vic2row = dict()
    for row in this.server.graphstudy_manager.sql_select(table, layout=layoutid):
        vic = row['vicinity']
        if vicinities is None or vic in vicinities:
            vic2row[vic] = row
    if not vic2row:
        raise HttpError(http.HTTPStatus.NOT_FOUND, "There is nothing to be plotted.")
    for vic in (set(vicinities) if vicinities is not None else set()) - set(vic2row.keys()):
        msg = "No density for property {:s} of layout {!s} with vicinity {:d} available.".format(
            prop.name, layoutid, vic)
        raise HttpError(http.HTTPStatus.NOT_FOUND, msg)
    xunit = PROPERTY_QUANTITIES[prop][0]
    colorfmt = '#{:06X}'
    (color1a, color1b) = (colorfmt.format(c) for c in LAYOUTS_FALLBACK_COLORS)
    color2 = '#{:06X}'.format(SECONDARY_COLOR)
    xlabel = "{0[1]:s} / {0[0]:s}".format(PROPERTY_QUANTITIES[prop])
    ylabel = "relative density / {:s}^{{-1}}".format(xunit)
    script = textwrap.dedent('''\
    #! {gp:s}
    #! -*- coding:utf-8; mode:gnuplot; -*-"

    set terminal {fmt:s} size 800,600
    set output
    set xrange [0 : *]
    set yrange [0 : *]
    set xlabel {xlabel!r}
    set ylabel {ylabel!r}
    set format y '%.2E'
    ''').format(gp=find_tool_lazily(GNUPLOT), fmt=terminal, xlabel=xlabel, ylabel=ylabel)
    plotexprs = list()
    if vicinities is None:
        for kv in sorted(vic2row.items(), key=get_first):
            datasource = _gpltin(kv[1]['file'])
            plotexprs.append("{df!r} with lines linecolor rgb {lc!r} notitle".format(df=datasource, lc=color1b))
    else:
        for (vic, row) in sorted(vic2row.items(), key=get_first):
            x, y = (min(row['maximum'], 0.9 * row['max-x']), 0.8 * row['max-y'])
            script += (
                "\n" +
                "set arrow from first {xl:G}, first {y:G} to first {xr:G}, first {y:G}"
                + " heads size graph 0.01, 90 front linecolor rgb {lc!r}\n" +
                "set label {label!r} at first {x:G}, first {y:G} center front offset first 0, graph 0.03\n"
            ).format(
                x=_GN(x), xl=_GN(x - row['sigma']), xr=_GN(x + row['sigma']), y=_GN(y),
                lc=color2, label="\u03C3 = \u00B1 {:,.3g} {:s}".format(row['sigma'], xunit)
            )
            fmtparams = {
                'df' : _gpltin(row['file']), 'lc1a' : color1a, 'lc1b' : color1b,
                'curve' : "{:,d} hops".format(vic),
                'points' : "{:,d} sampling points".format(row['points']),
            }
            plotexprs.append("{df!r} smooth csplines linecolor rgb {lc1b!r} title {curve!r}".format(**fmtparams))
            plotexprs.append("{df!r} with points     linecolor rgb {lc1a!r} title {points!r}".format(**fmtparams))
    assert plotexprs
    script += "\nplot " + ', \\\n     '.join(plotexprs)
    _serve_plot(this, script, mimetype, charset, source=serve_source)

def _serve_entropy(this, prop, graphid, url, format=None):
    kind = PropertyKinds.classify(prop)
    if kind not in { PropertyKinds.SCALED_HISTOGRAM, PropertyKinds.GLIDING_AVERAGE }:
        msg = "No entropy for property {:s} available.".format(prop.name)
        raise HttpError(http.HTTPStatus.NOT_FOUND, msg)
    (terminal, mimetype, charset) = _check_graphics_format(this, format)
    query = urllib.parse.parse_qs(url.query)
    serve_source = get_bool_from_query(query, 'source')
    query = urllib.parse.parse_qs(url.query)
    try:
        selected_layouts = frozenset(validate_layout_id(token) for token in query['layouts'])
    except KeyError:
        selected_layouts = None
    keyname = { PropertyKinds.SCALED_HISTOGRAM : 'bincount', PropertyKinds.GLIDING_AVERAGE  : 'vicinity' }[kind]
    entropies = dict()
    layouts = { l : None for l in Layouts }
    with this.server.graphstudy_manager.sql_ctx as curs:
        for layorow in this.server.graphstudy_manager.sql_select_curs(curs, 'Layouts', graph=graphid, layout=object):
            layoutid = layorow['id']
            if selected_layouts is not None and layoutid not in selected_layouts:
                continue
            entropy = dict()
            if kind is PropertyKinds.SCALED_HISTOGRAM:
                table = 'ScaledHistograms'
                kwargs = { 'property' : prop, 'layout' : layoutid, 'binning' : Binnings.FIXED_COUNT }
            else:
                table = SQL_PROPERTY_TABLES[prop]
                kwargs = { 'layout' : layoutid }
            for proprow in this.server.graphstudy_manager.sql_select_curs(curs, table, **kwargs):
                entropy[proprow[keyname]] = value_or(proprow['entropy'], math.nan)
            if not entropy:
                logging.warning("Cannot plot entropy for {:s} of layout {!s} (no data)".format(prop.name, layoutid))
            entropies[layoutid] = entropy
            layouts[layorow['layout']] = layoutid
    if selected_layouts is not None:
        for lid in selected_layouts - set(layouts.values()):
            raise HttpError(http.HTTPStatus.NOT_FOUND, "No layout with ID {!s} exists.".format(lid))
    if not any(entropies.values()):
        raise HttpError(http.HTTPStatus.NOT_FOUND, "There is nothing to be plotted.")
    script = [
        "#! {:s}".format(find_tool_lazily(GNUPLOT)),
        "#! -*- coding:utf-8; mode:gnuplot; -*-",
        "",
        "set terminal {:s} size 800,600".format(terminal),
        "set output",
        "set xrange [0 : *]",
        "set yrange [0 : *]",
        "set xlabel {!r}".format("log_2({:s})".format(keyname)),
        "set ylabel {!r}".format("entropy / bit"),
        "set format x '%.1f'",
        "set format y '%.2E'",
        "set key left top Left reverse noenhanced",
        "",
    ]
    plotexprs = list()
    for (lo, lid) in sorted(filter(get_second, layouts.items()), key=get_first):
        (d, k) = _linear_regression(list((math.log2(x), y) for (x, y) in entropies[lid].items()))
        script.append("{:20s} = {d:20G} + {k:20G} * x".format('f_' + lo.name.lower() + '(x)', d=_GN(d), k=_GN(k)))
        (color1st, color2nd) = ('#{:06X}'.format(c) for c in LAYOUTS_PREFERRED_COLORS[lo])
        title = '{:s} (d = {:,.3g}, k = {:,.3g})'.format(lo.name, d, k)
        plotexprs.append("{!r} using 1:2 linecolor rgb {!r} title {!r}".format('-', color1st, title))
        plotexprs.append("{:s}(x) linecolor rgb {!r} notitle".format('f_' + lo.name.lower(), color2nd))
    script.append("")
    script.append("plot " + ', \\\n     '.join(plotexprs))
    script.append("")
    for (lo, lid) in sorted(filter(get_second, layouts.items()), key=get_first):
        script.append("")
        script.append("# {!s} ({:s})".format(lid, lo.name))
        script.append("# {:>18s} {:>20s}".format("key", "entropy / bit"))
        for kv in sorted(entropies[lid].items(), key=get_first):
            script.append("{x:20G} {y:20G}".format(x=_GN(math.log2(kv[0])), y=_GN(kv[1])))
        script.append('e')
    script.append("")  # ensure trailing newline at end of file
    _serve_plot(this, '\n'.join(script), mimetype, charset, source=serve_source)

def _serve_plot(this, script, mimetype, charset, source=False):
    if source:
        data = script.encode('utf-8')
        this.send_response(http.HTTPStatus.OK)
        this.send_header('Content-Length', str(len(data)))
        this.send_header('Content-Type', 'text/plain; charset="UTF-8"')
        this.send_header('Content-Disposition', 'inline; filename="gnuplot.txt"')
        this.end_headers()
        this.wfile.write(data)
        this.wfile.flush()
    else:
        try:
            data = this.server.graphstudy_manager.call_gnuplot(script, stdout=True)
        except RecoverableError as e:
            raise HttpError(http.HTTPStatus.INTERNAL_SERVER_ERROR, "Cannot create plot: {!s}".format(e))
        else:
            this.send_response(http.HTTPStatus.OK)
            this.send_header('Content-Length', str(len(data)))
            this.send_header('Content-Type', '{:s}; charset="{:s}"'.format(mimetype, charset))
            this.end_headers()
            this.wfile.write(data)
            this.wfile.flush()

def _get_bincount_from_query(query):
    try:
        [ bcs ] = query['bincount']
    except KeyError:
        return None
    except ValueError:
        raise HttpError(http.HTTPStatus.BAD_REQUEST, "Zero or one 'bincount' parameters are allowed.")
    try:
        bc = int(bcs)
        if bc >= 0: return bc
    except ValueError:
        pass
    raise HttpError(http.HTTPStatus.BAD_REQUEST, "Parameter 'bincount' must be set to a non-negative integer.")

def _get_vicinities_from_query(query):
    vic = query.get('vicinity')
    if not vic or '*' in vic:
        return None
    try:
        vics = list(map(int, vic))
        if all(l >= 0 for l in vics): return vics
    except ValueError:
        pass
    raise HttpError(http.HTTPStatus.BAD_REQUEST, "Parameter 'vicinity' must be set to a non-negative integer.")

def _linear_regression(xydata):
    if len(xydata) < 3:
        return (math.nan, math.nan)
    xmean = statistics.mean(map(get_first, xydata))
    ymean = statistics.mean(map(get_second, xydata))
    try:
        k = sum((x - xmean) * (y - ymean) for (x, y) in xydata) / sum((x - xmean)**2 for (x, y) in xydata)
        d = ymean - k * xmean
        return (d, k)
    except ArithmeticError as e:
        logging.error("Cannot fit linear regression: {!s}".format(e))
        return (math.nan, math.nan)

def _check_graphics_format(this, fmt):
    if not isinstance(fmt, str):
        raise TypeError()
    elif fmt == 'SVG':
        return ('svg', 'image/svg+xml', 'binary')
    elif fmt == 'PNG':
        return ('png', 'image/png', 'binary')
    else:
        raise HttpError(http.HTTPStatus.NOT_FOUND, "The graphics format {!r} is not supported.".format(fmt))

def _gpltin(filename):
    ext = os.path.splitext(filename)[1].upper().lstrip('.')
    if ext in [ 'TXT', 'DAT' ]:
        return filename
    elif ext == 'GZ':
        return '<{:s} {:s}'.format(find_tool_lazily(GZIP_FILTER), filename)
    elif ext == 'BZ2':
        return '<{:s} {:s}'.format(find_tool_lazily(BZ2_FILTER), filename)
    else:
        raise HttpError(http.HTTPStatus.INTERNAL_SERVER_ERROR, "Cannot infer type of data file {!r}".format(filename))

class _UnknownHandler(object):

    def __init__(self, kind):
        self.__kind = kind

    def __call__(self, *args, **kwargs):
        msg = "The specified plot type {!r} is unknown to mankind.".format(self.__kind)
        raise HttpError(http.HTTPStatus.NOT_FOUND, msg)

class _Handler(object):

    def __init__(self):
        self.__handlers = {
            'ENTROPY' : _serve_entropy,
            'GLIDING' : _serve_gliding_average,
            'HISTO' : _serve_histogram,
        }

    def __getitem__(self, key):
        key = translate_enum_name(key, old='-', new='_', case='upper')
        try:
            return self.__handlers[key]
        except KeyError:
            return _UnknownHandler(key)

serve = _Handler()
