#! /usr/bin/python3
#! -*- coding:utf-8; mode:python; -*-

import argparse
import json
import sqlite3
import statistics
import subprocess
import sys

class Confusion(object):

    def __init__(self, tn, fn, fp, tp, total=None):
        count = tn + fn + fp + tp
        assert total is None or total == count
        self.test_count = count
        self.true_negative = tn / count
        self.false_negative = fn / count
        self.false_positive = fp / count
        self.true_positive = tp / count
        self.condition_negative = (tn + fp) / count
        self.condition_positive = (tp + fn) / count
        self.prediction_negative = (tn + fn) / count
        self.prediction_positive = (tp + fp) / count
        self.success = (tn + tp) / count
        self.failure = (fn + fp) / count

def main():
    ap = argparse.ArgumentParser(
        description="Perfoms Monte Carlo cross validation and outputs a confusion matrix with estimated errors"
    )
    ap.add_argument('-r', '--runs', metavar='N', type=int, required=True, help="repeat N times")
    ap.add_argument('-d', '--database', metavar='FILE', required=True, help="communicate via SQLite DB in FILE")
    ap.add_argument('-o', '--output', metavar='FILE', help="write confusion matrix to FILE")
    ap.add_argument('-q', '--quiet', action='store_true', help="discard driver logging output")
    ap.add_argument('cmd', nargs='+', help="command to execute for rebuilding and -testing the model")
    ns = ap.parse_args()
    result = cross_validate(ns.cmd, ns.database, ns.runs, quiet=ns.quiet)
    def writeout(stream):
        json.dump(result, stream, sort_keys=True, indent=4)
        stream.write('\n')
    if ns.output is None or ns.output == '-':
        writeout(sys.stdout)
    else:
        with open(ns.output, 'w') as ostr:
            writeout(ostr)

def cross_validate(command, database, runs, quiet=False):
    results = list()
    for i in range(runs):
        print("Running corss validation {:d} / {:d} ...".format(i + 1, runs), file=sys.stderr)
        stderr = subprocess.DEVNULL if quiet else None
        subprocess.run(command, check=True, stdin=subprocess.DEVNULL, stdout=subprocess.DEVNULL, stderr=stderr)
        results.append(get_confusion(database))
    return combine_confusions(results)

def get_confusion(database):
    conn = None
    try:
        conn = sqlite3.connect(database)
        conn.row_factory = sqlite3.Row
        with conn as curs:
            basequery = "SELECT COUNT() FROM `NNTestScore`"
            total = get_count_from_query(curs, basequery)
            tn = get_count_from_query(curs, basequery + " WHERE `expected` < 0.0 AND `actual` < 0.0")
            fn = get_count_from_query(curs, basequery + " WHERE `expected` > 0.0 AND `actual` < 0.0")
            fp = get_count_from_query(curs, basequery + " WHERE `expected` < 0.0 AND `actual` > 0.0")
            tp = get_count_from_query(curs, basequery + " WHERE `expected` > 0.0 AND `actual` > 0.0")
    finally:
        if conn is not None:
            conn.close()
    return Confusion(tn=tn, fn=fn, fp=fp, tp=tp, total=total)

def get_count_from_query(curs, query, params=None):
    rows = curs.execute(query, params if params is not None else tuple()).fetchall()
    assert len(rows) == 1
    assert len(rows[0]) == 1
    count = rows[0][0]
    assert type(count) is int
    return count

def combine_confusions(confusions):
    info = dict()
    info['test-runs'] = len(confusions)
    attributes = [
        'test-count', 'success', 'failure',
        'true-negative', 'false-negative', 'false-positive', 'true-positive',
        'condition-positive', 'condition-negative',
        'prediction-positive', 'prediction-negative',
    ]
    for what in attributes:
        (mean, stdev) = make_statistic(confusions, what.replace('-', '_'))
        info[what] = { 'mean' : mean, 'stdev' : stdev }
    return info

def make_statistic(confusions, key):
    values = [ getattr(c, key) for c in confusions ]
    mean = statistics.mean(values)
    stdev = statistics.stdev(values, xbar=mean)
    return (mean, stdev)

def get_one(seq):
    [ x ] = seq
    return x

if __name__ == '__main__':
    main()
