/* -*- coding:utf-8; mode:c++; -*- */

/**
 * @file cli.hxx
 *
 * @brief
 *     Common facilities for building CLIs.
 *
 */

#ifndef MSC_CLI_HXX
#define MSC_CLI_HXX

#include <cstddef>
#include <iosfwd>
#include <map>
#include <optional>
#include <string>
#include <string_view>
#include <vector>

#include <boost/program_options.hpp>

#include "enums/kernels.hxx"
#include "file.hxx"

namespace msc
{

    /**
     * @brief
     *     Base-class for command-line interfaces.
     *
     * All strings must not contain vertical white-space and should only contain simple text.
     *
     */
    struct cli_base
    {
        /** @brief Informal name of the program.  */
        std::string prog{};

        /** @brief Program invocation synopsis.  */
        std::string usage{};

        /** @brief Zero or more paragraphs of general information about the program.  */
        std::vector<std::string> help{};

        /** @brief Zero or more paragraphs of final remarks about the program.  */
        std::vector<std::string> epilog{};

        /** @brief Description of influential environment variables.  */
        std::map<std::string, std::string> environ{};
    };

    /**
     * @brief
     *     Generic command-line interface.
     *
     * In order to build a CLI using this facility, you need to define a `struct` that has (at least) all well-known
     * command-line parameters as members.  The following members will be understood.  The `struct` can be named
     * anything but let's call it `ParamsT` for now.
     *
     * <table>
     *   <tr>
     *     <th>flag</th>
     *     <th>option</th>
     *     <th>member</th>
     *     <th>type</th>
     *     <th>default</th>
     *     <th>comment</th>
     *   </tr>
     *   <tr>
     *     <td>&nbsp;</td>
     *     <td></td>
     *     <td>`input`</td>
     *     <td>`std::string`</td>
     *     <td>`"-"`</td>
     *     <td>optional positional argument</td>
     *   </tr>
     *   <tr>
     *     <td>&nbsp;</td>
     *     <td></td>
     *     <td>`input1st`</td>
     *     <td>`input_file`</td>
     *     <td>`"-"`</td>
     *     <td>mandatory positional argument</td>
     *   </tr>
     *   <tr>
     *     <td>&nbsp;</td>
     *     <td></td>
     *     <td>`input2nd`</td>
     *     <td>`input_file`</td>
     *     <td>`"-"`</td>
     *     <td>mandatory positional argument</td>
     *   </tr>
     *   <tr>
     *     <td>`-o`</td>
     *     <td>`--output`</td>
     *     <td>`output`</td>
     *     <td>`output_file`</td>
     *     <td>`"-"`</td>
     *     <td>optional</td>
     *   </tr>
     *   <tr>
     *     <td>`-m`</td>
     *     <td>`--meta`</td>
     *     <td>`meta`</td>
     *     <td>`output_file`</td>
     *     <td>`""`</td>
     *     <td>optional</td>
     *   </tr>
     *   <tr>
     *     <td>`-f`</td>
     *     <td>`--format`</td>
     *     <td>`format`</td>
     *     <td>`#fileformats`</td>
     *     <td><em>see comment</em></td>
     *     <td>mandatory if initialized to zero, otherwise optional with that default</td>
     *   </tr>
     *   <tr>
     *     <td></td>
     *     <td>`--list-formats`</td>
     *     <td></td>
     *     <td></td>
     *     <td></td>
     *     <td>always added together with `--format`</td>
     *   </tr>
     *   <tr>
     *     <td>`-l`</td>
     *     <td>`--layout`</td>
     *     <td>`layout`</td>
     *     <td>`bool`</td>
     *     <td>`false`</td>
     *     <td>boolean flag</td>
     *   </tr>
     *   <tr>
     *     <td>`-y`</td>
     *     <td>`--simplify`</td>
     *     <td>`simplify`</td>
     *     <td>`bool`</td>
     *     <td>`false`</td>
     *     <td>boolean flag</td>
     *   </tr>
     *   <tr>
     *     <td>`-n`</td>
     *     <td>`--nodes`</td>
     *     <td>`nodes`</td>
     *     <td>`int`</td>
     *     <td>0 &lt; <var>N</var> &le; `INT_MAX`</td>
     *     <td>optional</td>
     *   </tr>
     *   <tr>
     *     <td>`-t`</td>
     *     <td>`--torus`</td>
     *     <td>`torus`</td>
     *     <td>`int`</td>
     *     <td><var>N</var> &ge; 0</td>
     *     <td>optional</td>
     *   </tr>
     *   <tr>
     *     <td>`-h`</td>
     *     <td>`--hyperdim`</td>
     *     <td>`hyperdim`</td>
     *     <td>`int`</td>
     *     <td><var>N</var> &gt; 0</td>
     *     <td>optional</td>
     *   </tr>
     *   <tr>
     *     <td>`-s`</td>
     *     <td>`--symmetric`</td>
     *     <td>`symmetric`</td>
     *     <td>`bool`</td>
     *     <td>`false`</td>
     *     <td>boolean flag</td>
     *   </tr>
     *   <tr>
     *     <td>`-a`</td>
     *     <td>`--algorithm`</td>
     *     <td>`algorithm`</td>
     *     <td>`#algorithms`</td>
     *     <td><em>see comment</em></td>
     *     <td>mandatory if initialized to zero, otherwise optional with that default</td>
     *   </tr>
     *   <tr>
     *     <td></td>
     *     <td>`--list-algorithms`</td>
     *     <td></td>
     *     <td></td>
     *     <td></td>
     *     <td>always added together with `--algorithm`</td>
     *   </tr>
     *   <tr>
     *     <td>`-d`</td>
     *     <td>`--distribution`</td>
     *     <td>`distribution`</td>
     *     <td>`#distributions`</td>
     *     <td><em>see comment</em></td>
     *     <td>mandatory if initialized to zero, otherwise optional with that default</td>
     *   </tr>
     *   <tr>
     *     <td></td>
     *     <td>`--list-distributions`</td>
     *     <td></td>
     *     <td></td>
     *     <td></td>
     *     <td>always added together with `--distribution`</td>
     *   </tr>
     *   <tr>
     *     <td>`-j`</td>
     *     <td>`--projection`</td>
     *     <td>`projection`</td>
     *     <td>`#projections`</td>
     *     <td><em>see comment</em></td>
     *     <td>mandatory if initialized to zero, otherwise optional with that default</td>
     *   </tr>
     *   <tr>
     *     <td></td>
     *     <td>`--list-projections`</td>
     *     <td></td>
     *     <td></td>
     *     <td></td>
     *     <td>always added together with `--projection`</td>
     *   </tr>
     *   <tr>
     *     <td>`-r`</td>
     *     <td>`--rate`</td>
     *     <td>`rate`</td>
     *     <td>`std::vector<double>`</td>
     *     <td>empty</td>
     *     <td>zero or more values in the unit interval</td>
     *   </tr>
     *   <tr>
     *     <td>`-k`</td>
     *     <td>`--kernel`</td>
     *     <td>`kernel`</td>
     *     <td>`#kernels`</td>
     *     <td><em>see comment</em></td>
     *     <td>mandatory if initialized to zero, otherwise optional with that default</td>
     *   </tr>
     *   <tr>
     *     <td></td>
     *     <td>`--list-kernels`</td>
     *     <td></td>
     *     <td></td>
     *     <td></td>
     *     <td>always added together with `--kernel`</td>
     *   </tr>
     *   <tr>
     *     <td>`-w`</td>
     *     <td>`--width`</td>
     *     <td>`width`</td>
     *     <td>`std::vector&lt;double&gt;`</td>
     *     <td>empty</td>
     *     <td>may be passed zero or more times</td>
     *   </tr>
     *   <tr>
     *     <td>`-b`</td>
     *     <td>`--bins`</td>
     *     <td>`bins`</td>
     *     <td>`std::vector&lt;int&gt;`</td>
     *     <td>empty</td>
     *     <td>may be passed zero or more times</td>
     *   </tr>
     *   <tr>
     *     <td>`-p`</td>
     *     <td>`--points`</td>
     *     <td>`points`</td>
     *     <td>`std::optional&lt;int&gt;`</td>
     *     <td>`std::nullopt`</td>
     *     <td>optional</td>
     *   </tr>
     *   <tr>
     *     <td>`-1`</td>
     *     <td>`--major`</td>
     *     <td>`component`</td>
     *     <td>`int`</td>
     *     <td>`0`</td>
     *     <td>stores `1` in `component` when passed</td>
     *   </tr>
     *   <tr>
     *     <td>`-2`</td>
     *     <td>`--minor`</td>
     *     <td>`component`</td>
     *     <td>`int`</td>
     *     <td>`0`</td>
     *     <td>stores `2` in `component` when passed</td>
     *   </tr>
     *   <tr>
     *     <td>`-c`</td>
     *     <td>`--clever`</td>
     *     <td>`clever`</td>
     *     <td>`bool`</td>
     *     <td>`false`</td>
     *     <td>boolean flag</td>
     *   </tr>
     *   <tr>
     *     <td></td>
     *     <td>`--node-color`</td>
     *     <td>`node_color`</td>
     *     <td>`ogdf::Color`</td>
     *     <td>anything</td>
     *     <td>optional</td>
     *   </tr>
     *   <tr>
     *     <td></td>
     *     <td>`--edge-color`</td>
     *     <td>`edge_color`</td>
     *     <td>`ogdf::Color`</td>
     *     <td>anything</td>
     *     <td>optional</td>
     *   </tr>
     *   <tr>
     *     <td></td>
     *     <td>`--tikz`</td>
     *     <td>`tikz`</td>
     *     <td>`bool`</td>
     *     <td>`false`</td>
     *     <td>boolean flag</td>
     *   </tr>
     *   <tr>
     *     <td></td>
     *     <td>`--help`</td>
     *     <td></td>
     *     <td></td>
     *     <td></td>
     *     <td>always added</td>
     *   </tr>
     *   <tr>
     *     <td></td>
     *     <td>`--version`</td>
     *     <td></td>
     *     <td></td>
     *     <td></td>
     *     <td>always added</td>
     *   </tr>
     * </table>
     *
     * Next, you have to write a `struct` (let's call it `AppT`) with a member of type `ParamsT` that must be named
     * `parameters`.  Furthermore, the `struct` must have its call operator overloaded to take no arguments and return
     * `void`.  (Errors should be reported via exceptions.)
     *
     * @tparam AppT
     *     see above
     *
     */
    template <typename AppT>
    class command_line_interface final : public cli_base
    {

        static_assert(
            std::is_same_v<void, decltype(std::declval<AppT&>()())>,
            "AppT must have an overloaded call operator that takes no arguments and returns `void`.  Errors have to be"
            " reported via exceptions."
        );

        static_assert(
            std::is_class_v<decltype(AppT::parameters)>,
            "AppT must define a member `parameters` for holding the parsed command-line parameters of the application."
        );

    public:

        /**
         * @brief
         *     Creates a new command-line application.
         *
         * @param prog
         *     informal program name
         *
         * @param app
         *     application to run internally
         *
         */
        explicit command_line_interface(std::string prog, AppT app = AppT{});

        /**
         * @brief
         *     Runs the main function `AppT::operator()` in a `try` / `catch` block after command-line parsing and quite
         *     a bit more.
         *
         * Here is how this function proceeds:
         *
         * 1. Parse the given command-line arguments and use them to initialize the members of `AppT::parameters` as
         *    described above.  If an option such as `--help` is encountered, a sythesized text using the attributes
         *    from the `cli_base` base (which therefore should be initialized) is shown and the following steps are
         *    skipped.
         *
         * 2. Call `set_resource_limits()`.  This is done after command-line parsing so that if there is an error
         *    setting the resource limits, users can at least view the help message (and therefore maybe figure out what
         *    to change to avoid the error).  It is highly unlikely that command-line parsing would exceed any
         *    reasonable limit.
         *
         * 3. Invoke `AppT::operator()`.
         *
         * 4. If step (1) through (3) raised any exceptions, handle them by printing an appropriate error message.
         *
         * 5. Test if standard input (`std::cin`) or standard output (`std::cout`)  are okay.
         *
         * 6. Return `EXIT_FAILURE` if an invalid command-line was passed, an exception thrown or the check in (5)
         *    failed and `EXIT_SUCCESS` otherwise.
         *
         * Any I/O done by the internal application nonwithstanding , all I/O generated by this class directly will go
         * to the standard C++ streams `std::cout` and `std::cerr`.  Low-level C or POSIX I/O will not be used.
         *
         * @param argc
         *     number of arguments as specified by the C standard for `main`'s first argument
         *
         * @param argv
         *     argument vector as specified by the C standard for `main`'s second argument
         *
         * @returns
         *     exit status for the application
         *
         */
        [[nodiscard]] int operator()(const int argc, const char *const *const argv);

        /**
         * @brief
         *     Returns a pointer to the internal application instance.
         *
         * @note
         *     The sole reason for this function is unit testing.
         *
         * @returns
         *     pointer to internal application
         *
         */
        AppT* operator->() noexcept
        {
            return &_app;
        }

    private:

        /** @brief The internal application.  */
        AppT _app{};

    };  // class command_line_interface

    /**
     * @brief
     *     Guesses the width of the output terminal.
     *
     * If the environment variable `COLUMNS` is set and can be parsed as an integer with a value in the interval
     * `(0, INT_MAX]`, that value is returned.  Otherwise, `fallback` is returned.
     *
     * @param fallback
     *     value to return as a last resort
     *
     * @returns
     *     guestimated terminal width in columns
     *
     */
    int guess_terminal_width(int fallback = 80) noexcept;

    /**
     * @brief
     *     Checks whether standard input or output are in an error state.
     *
     * EOF is not considered an error state.
     *
     * The intended use of this function is to be called at the end of `main` as `check_stdio(std::cin, std::cout)` to
     * make sure I/O errors are not silently passed over.
     *
     * @param input
     *     standard input stream
     *
     * @param output
     *     standard output stream
     *
     * @throws std::system_error
     *     if either stream is not okay
     *
     */
    void check_stdio(std::istream& input, std::ostream& output);

    /**
     * @brief
     *     Checks whether standard input or output are in an error state.
     *
     * This function performs the same check as the throwing overload but instead of throwing an exception returns
     * whether the streams were both okay and, if they are not, provides an appropriate error message.
     *
     * In the rare event that assigning to the string required a memory reallocation that fails, `std::terminate` will
     * be called.  If this is a concern, make sure to `reserve()` sufficient space in the string beforehand.
     *
     * @param input
     *     standard input stream
     *
     * @param output
     *     standard output stream
     *
     * @param message
     *     if `false` is returned, an error message is assigned to the string, otherwise, it is left unchanged
     *
     * @returns
     *     whether both streams are okay
     *
     */
    bool check_stdio(std::istream& input, std::ostream& output, std::string& message) noexcept;

    const std::string& helptext_file_name_expansion();

    struct cli_parameters_interpolation
    {
        input_file input1st{};
        input_file input2nd{};
        output_file output{"-"};
        output_file meta{};
        bool clever{};
        std::vector<double> rate{};

        output_file expand_filename(double degree) const;
    };

    struct cli_parameters_worsening
    {
        input_file input{"-"};
        output_file output{"-"};
        output_file meta{};
        std::vector<double> rate{};

        output_file expand_filename(double degree) const;
    };

    struct cli_parameters_property
    {
        input_file input{"-"};
        output_file output{"-"};
        output_file meta{};
        msc::kernels kernel{};
        std::vector<double> width{};
        std::vector<int> bins{};
        std::optional<int> points{};

        std::size_t iterations() const noexcept;
        output_file expand_filename(std::size_t iteration) const;
    };

    struct cli_parameters_metric
    {
        input_file input{"-"};
        //output_file output{"-"};
        output_file meta{};
    };

}  // namespace msc

#define MSC_INCLUDED_FROM_CLI_HXX
#include "cli.txx"
#undef MSC_INCLUDED_FROM_CLI_HXX

#endif  // !defined(MSC_CLI_HXX)
