/* -*- coding:utf-8; mode:c++; -*- */

/**
 * @file cuboid.hxx
 *
 * @brief
 *     <var>N</var>-dimensional cuboids and related functions.
 *
 */

#ifndef MSC_CUBOID_HXX
#define MSC_CUBOID_HXX

#include <array>
#include <cassert>
#include <cstdint>
#include <iterator>
#include <memory>
#include <type_traits>
#include <utility>

#include "ogdf_fwd.hxx"
#include "point.hxx"

namespace msc
{

    /**
     * @brief
     *     An axis-aligned <var>N</var>-dimensional cuboid.
     *
     */
    template <typename T, std::size_t N>
    class cuboid final
    {

        static_assert(std::is_floating_point_v<T>, "Cuboids can only have real coordinates");
        static_assert((N >= 2) && (N <= 30), "Cuboids can only have 2 ... 30 dimensions");

    public:

        using point_type = point<T, N>;
        using index_type = std::uint32_t;

        constexpr cuboid() noexcept = default;

        constexpr cuboid(const point_type& org, const point_type& ext) noexcept : _origin{org}, _extension{ext}
        {
        }

        static constexpr std::size_t dimensions() noexcept;

        static constexpr std::size_t corners() noexcept;

        static constexpr std::array<index_type, N> neighbours(index_type idx);

        constexpr point_type corner(index_type idx) const;

        constexpr const point_type& origin() const noexcept { return _origin; }

        constexpr point_type& origin() noexcept { return _origin; }

        constexpr const point_type& extension() const noexcept { return _extension; }

        constexpr point_type& extension() noexcept { return _extension; }

    private:

        point<T, N> _origin{};
        point<T, N> _extension{};

    };  // cuboid

    template
    <
        typename FwdIterT,
        typename CuboidT = typename std::iterator_traits<FwdIterT>::value_type,
        typename PointT = typename CuboidT::point_type,
        typename T = typename PointT::value_type,
        std::size_t N = std::tuple_size<PointT>::value
    >
    std::pair<std::unique_ptr<ogdf::Graph>, std::unique_ptr<ogdf::GraphAttributes>>
    convert_and_project(const FwdIterT first, const FwdIterT last);

}  // namespace msc

#define MSC_INCLUDED_FROM_CUBOID_HXX
#include "cuboid.txx"
#undef MSC_INCLUDED_FROM_CUBOID_HXX

#endif  // !defined(MSC_CUBOID_HXX)
