/* -*- coding:utf-8; mode:c++; -*- */

#ifndef MSC_INCLUDED_FROM_DATA_ANALYSIS_HXX
#  error "Never `#include <data_analysis.txx>` directly; `#include <data_analysis.hxx>` instead"
#endif

#include <algorithm>
#include <cassert>
#include <cfloat>
#include <cmath>
#include <utility>
#include <vector>

#include "gliding.hxx"
#include "histogram.hxx"
#include "io.hxx"
#include "logging.hxx"
#include "stochastic.hxx"
#include "useful.hxx"

namespace msc
{

    namespace data_analysis_detail
    {

        template <typename FwdIterT>
        constexpr bool _distance_less_than_three(const FwdIterT first, const FwdIterT last) noexcept
        {
            // We could implement this function as `return std::distance(first, last) < 3;` but it would be an O(N)
            // operation which we cannot afford for large N.
            auto tally = 0;
            for (auto it = first; it != last; ++it) {
                if (++tally == 3) {
                    return false;
                }
            }
            return true;
        }

    }  // namespace data_analysis_detail

    inline data_analyzer::data_analyzer(const kernels kern) noexcept : _kernel{kern}
    {
        assert(kern != kernels{});
    }

    template <typename ContainerT>
    void data_analyzer::analyze(const ContainerT& sample, json_object& info, json_object& subinfo) const
    {
        if (!this->analyze_oknodo(sample, info, subinfo)) {
            throw std::logic_error{"Not enough data for a statistical analysis"};
        }
    }

    template <typename FwdIterT>
    void data_analyzer::analyze(const FwdIterT first,
                                const FwdIterT last,
                                json_object& info,
                                json_object& subinfo) const
    {
        if (!this->analyze_oknodo(first, last, info, subinfo)) {
            throw std::logic_error{"Not enough data for a statistical analysis"};
        }
    }

    template <typename ContainerT>
    [[nodiscard]] bool data_analyzer::analyze_oknodo(const ContainerT& sample,
                                                     json_object& info,
                                                     json_object& subinfo) const
    {
        return this->analyze_oknodo(std::begin(sample), std::end(sample), info, subinfo);
    }

    template <typename FwdIterT>
    [[nodiscard]] bool data_analyzer::analyze_oknodo(const FwdIterT first,
                                                     const FwdIterT last,
                                                     json_object& info,
                                                     json_object& subinfo) const
    {
        // TODO: It would be so much better to split this into a switch statement with a single return per case that
        //       calls a dedicated subroutine to perform the actual work.
        if (data_analysis_detail::_distance_less_than_three(first, last)) {
            MSC_LOG_NOTICE("Doing nothing for sample with less than three events");
            return false;
        } else if (_kernel == kernels::boxed) {
            const auto histo = [width = _width, count = optional_cast<std::size_t>(_bins)](const auto f, const auto l){
                if      (!width && !count) return histogram{f, l};
                else if ( width && !count) return histogram{f, l, *width};
                else if (!width &&  count) return histogram{f, l, *count};
                else throw std::invalid_argument{"Cannot fix bin width and bin count at the same time"};
            }(first, last);
            write_frequencies(histo, _output);
            _update_info(info, subinfo, histo);
        } else if (_kernel == kernels::gaussian) {
            const auto summary = get_stochastic_summary(first, last);
            const auto lo = _lower.value_or(summary.min);
            const auto hi = _upper.value_or(summary.max);
            if (lo > hi) {
                MSC_LOG_NOTICE("Invalid range [%g, %g]", lo, hi);
                return false;
            }
            const auto scott = (summary.mean == summary.rms)
                ? 1.0
                : binwidth_scott_normal_reference(summary.count, summary.stdev());
            const auto other = _bins.has_value()
                ? std::make_optional(((lo < hi) && (*_bins > 1)) ? (hi - lo) / (*_bins - 1) : 1.0)
                : std::nullopt;
            const auto sigma = _width.value_or(other.value_or(scott)) / 2.0;
            if (sigma <= 0.0) {
                throw std::invalid_argument{"sigma must be positive"};
            }
            const auto filter = gaussian_kernel{first, last, sigma};
            const auto delta = 3.0 * sigma;
            const auto density = _points
                ? make_density(filter, lo - delta, hi + delta, *_points)
                : make_density_adaptive(filter, lo - delta, hi + delta);
            write_density(density, summary, _output);
            const auto entropy = get_differential_entropy_of_pdf(density);
            const auto compx = [](auto&& lhs, auto&& rhs){ return lhs.first  < rhs.first; };
            const auto compy = [](auto&& lhs, auto&& rhs){ return lhs.second < rhs.second; };
            const auto maxx = std::max_element(std::begin(density), std::end(density), compx)->first;
            const auto maxy = std::max_element(std::begin(density), std::end(density), compy)->second;
            _update_info(info, subinfo, summary, sigma, density.size(), entropy, std::make_pair(maxx, maxy));
        } else if (_kernel == kernels::raw) {
            const auto events = std::vector<double>{first, last};
            const auto summary = get_stochastic_summary(events);
            write_events(events, summary, _output);
            _update_info(info, subinfo, summary);
        } else {
            reject_invalid_enumeration(_kernel, "msc::kernels");
        }
        return true;
    }

    inline kernels data_analyzer::get_kernel() const noexcept
    {
        return _kernel;
    }

    inline void data_analyzer::set_kernel(const kernels kern)
    {
        assert(kern != kernels{});
        _kernel = kern;
    }

    inline std::optional<double> data_analyzer::get_lower() const noexcept
    {
        return _lower;
    }

    inline std::optional<double> data_analyzer::get_upper() const noexcept
    {
        return _upper;
    }

    inline void data_analyzer::set_range() noexcept
    {
        _lower = std::nullopt;
        _upper = std::nullopt;
    }

    inline void data_analyzer::set_range(const std::optional<double> lower, const std::optional<double> upper)
    {
        assert(!lower || std::isfinite(*lower));
        assert(!upper || std::isfinite(*upper));
        assert(lower.value_or(DBL_MIN) <= upper.value_or(DBL_MAX));
        _lower = lower;
        _upper = upper;
    }

    inline std::optional<double> data_analyzer::get_width() const noexcept
    {
        return _width;
    }

    inline void data_analyzer::set_width(const std::optional<double> width)
    {
        assert(!width || (std::isfinite(*width) && (*width > 0.0)));
        _width = width;
    }

    inline std::optional<int> data_analyzer::get_bins() const noexcept
    {
        return _bins;
    }

    inline void data_analyzer::set_bins(const std::optional<int> bins)
    {
        assert(!bins || *bins > 0);
        _bins = bins;
    }

    inline std::optional<int> data_analyzer::get_points() const noexcept
    {
        return _points;
    }

    inline void data_analyzer::set_points(const std::optional<int> points)
    {
        assert(!points || *points > 0);
        _points = points;
    }

    inline const output_file& data_analyzer::get_output() const noexcept
    {
        return _output;
    }

    inline void data_analyzer::set_output(output_file dst)
    {
        _output = std::move(dst);
    }

}  // namespace msc
