/* -*- coding:utf-8; mode:c++; -*- */

/**
 * @file gliding.hxx
 *
 * @brief
 *     Gliding averages (kernels).
 *
 */

#ifndef MSC_GLIDING_HXX
#define MSC_GLIDING_HXX

#include <type_traits>
#include <utility>
#include <vector>

#include "stochastic.hxx"

namespace msc
{

    template
    <
        typename FwdIterT,
        typename ValueT = typename std::iterator_traits<FwdIterT>::value_type,
        typename = std::enable_if_t<std::is_floating_point_v<ValueT>>
    >
    class gaussian_kernel final
    {
    public:

        gaussian_kernel(const FwdIterT first, const FwdIterT last, const ValueT sigma) :
            _first{first}, _last{last}, _sigma{sigma}
        {
        }

        ValueT operator()(const ValueT x) const
        {
            const auto kernel = gaussian{x, _sigma};
            auto accu = ValueT{};
            for (auto it = _first; it != _last; ++it) {
                accu += kernel(*it);
            }
            return accu;
        }

    private:

        FwdIterT _first{};
        FwdIterT _last{};
        ValueT _sigma{};

    };  // class gaussian_kernel

    template <typename KernelT>
    std::vector<std::pair<double, double>>
    make_density(const KernelT& kernel, double minval, double maxval, int points);

    template <typename KernelT>
    std::vector<std::pair<double, double>>
    make_density_adaptive(const KernelT& kernel, double minval, double maxval, bool normalize = true);

    /**
     * @brief
     *     Computes the differential entropy of a probability density function.
     *
     * The behavior is undefined if the density is not normalized.
     *
     * @param density
     *     pointwise probability density function
     *
     * @returns
     *     entropy of density
     *
     * @todo
     *     This is not a correct entropy measure (https://en.wikipedia.org/wiki/Differential_entropy).
     *
     */
    double get_differential_entropy_of_pdf(const std::vector<std::pair<double, double>>& density);

}  // namespace msc

#define MSC_INCLUDED_FROM_GLIDING_HXX
#include "gliding.txx"
#undef MSC_INCLUDED_FROM_GLIDING_HXX

#endif  // !defined(MSC_GLIDING_HXX)
