/* -*- coding:utf-8; mode:c++; -*- */

/**
 * @file stochastic.hxx
 *
 * @brief
 *     Common stochasitc functions.
 *
 */

#ifndef MSC_STOCHASTIC_HXX
#define MSC_STOCHASTIC_HXX

#include <algorithm>
#include <cassert>
#include <cmath>
#include <iterator>
#include <limits>
#include <numeric>
#include <type_traits>
#include <utility>

#include "useful.hxx"  // for square

namespace msc
{

    template <typename FwdIterT, typename ValueT = typename std::iterator_traits<FwdIterT>::value_type>
    std::enable_if_t<std::is_arithmetic_v<ValueT>, std::pair<ValueT, ValueT>>
    min_max(const FwdIterT first, const FwdIterT last)
    {
        assert(first != last);
        const auto [minit, maxit] = std::minmax_element(first, last);
        assert((minit != last) && (maxit != last));
        return {*minit, *maxit};
    }

    template <typename ContainerT, typename ValueT = typename ContainerT::value_type>
    std::enable_if_t<std::is_arithmetic_v<ValueT>, std::pair<ValueT, ValueT>>
    min_max(const ContainerT& container)
    {
        return min_max(std::begin(container), std::end(container));
    }

    template <typename FwdIterT, typename ValueT = typename std::iterator_traits<FwdIterT>::value_type>
    std::enable_if_t<std::is_floating_point_v<ValueT>, std::pair<ValueT, ValueT>>
    mean_stdev(const FwdIterT first, const FwdIterT last)
    {
        constexpr auto zero = ValueT{};
        const auto n = std::distance(first, last);
        assert(n >= 3);
        const auto mean = std::accumulate(first, last, zero) / n;
        const auto lambda = [mean](auto acc, auto x){ return acc + ::msc::square(x - mean); };
        const auto stdev = std::sqrt(std::accumulate(first, last, zero, lambda) / (n - 1));
        return {mean, stdev};
    }

    template <typename ContainerT, typename ValueT = typename ContainerT::value_type>
    std::enable_if_t<std::is_floating_point_v<ValueT>, std::pair<ValueT, ValueT>>
    mean_stdev(const ContainerT& container)
    {
        return mean_stdev(std::begin(container), std::end(container));
    }

    struct stochastic_summary final
    {

        std::size_t count{};
        double min{};
        double max{};
        double mean{};
        double rms{};

        constexpr stochastic_summary() noexcept = default;

        constexpr stochastic_summary(const std::size_t count,
                                     const double min,
                                     const double max,
                                     const double mean,
                                     const double rms) noexcept :
            count{count},
            min{min},
            max{max},
            mean{mean},
            rms{rms}
        {
        }

        constexpr stochastic_summary(std::nullptr_t) noexcept :
            min{std::numeric_limits<double>::quiet_NaN()},
            max{std::numeric_limits<double>::quiet_NaN()},
            mean{std::numeric_limits<double>::quiet_NaN()},
            rms{std::numeric_limits<double>::quiet_NaN()}
        {
        }

        double stdev() const noexcept
        {
            const auto n = static_cast<double>(count);
            return std::sqrt(n / (n - 1.0) * (square(this->rms) - square(this->mean)));
        }

        double stdevp() const noexcept
        {
            return std::sqrt(square(this->rms) - square(this->mean));
        }

    };  // struct stochastic_summary

    /**
     * @brief
     *     Computes numer of events, minumum, maximum, mean and RMS of a population given by a pair of forward
     *     iterators.
     *
     * The range is iterated only once and each iterator is dereferenced only once, which is preferable if iterating is
     * non-trivial.
     *
     * If the range is empty, the behavior is undefined.
     *
     * @tparam FwdIterT
     *     forward iterator type that can be dereferenced to a floating-point value
     *
     * @param first
     *     begin of the range
     *
     * @param last
     *     end of the range
     *
     * @returns
     *     basic stochastic properties of the population
     *
     */
    template <typename FwdIterT, typename ValueT = typename std::iterator_traits<FwdIterT>::value_type>
    std::enable_if_t<std::is_floating_point_v<ValueT>, stochastic_summary>
    get_stochastic_summary(const FwdIterT first, const FwdIterT last)
    {
        auto sum = ValueT{};
        auto sqrsum = ValueT{};
        auto min = std::numeric_limits<ValueT>::max();
        auto max = std::numeric_limits<ValueT>::min();
        auto count = std::size_t{};
        for (auto it = first; it != last; ++it) {
            const auto value = *it;
            sum += value;
            sqrsum += square(value);
            min = std::min(min, value);
            max = std::max(max, value);
            count += 1;
        }
        return {count, min, max, sum / count, std::sqrt(sqrsum / count)};
    }

    /**
     * @brief
     *     Computes numer of events, minumum, maximum, mean and RMS of a population given by a container.
     *
     * This is a not-so-useful convenience overload of the function that takes a pair of iterators.
     *
     * If the range is empty, the behavior is undefined.
     *
     * @tparam ContainerT
     *     container type holding floating-point values
     *
     * @param container
     *     container of events
     *
     * @returns
     *     basic stochastic properties of the population
     *
     */
    template <typename ContainerT, typename ValueT = typename ContainerT::value_type>
    std::enable_if_t<std::is_floating_point_v<ValueT>, stochastic_summary>
    get_stochastic_summary(const ContainerT& container)
    {
        return get_stochastic_summary(std::begin(container), std::end(container));
    }

    template <typename FwdIterT, typename ValueT = typename std::iterator_traits<FwdIterT>::value_type>
    std::enable_if_t<std::is_floating_point_v<ValueT>, ValueT>
    entropy(const FwdIterT first, const FwdIterT last)
    {
        constexpr auto zero = ValueT{0};
        assert(std::all_of(first, last, [](const auto x){ return (x >= zero); }));
        if (first == last) {
            return zero;
        }
        assert(std::abs(1.0 - std::accumulate(first, last, zero)) < 1.0e-6);
        const auto lambda = [](const auto acc, const auto x){
            return (x > zero) ? acc + x * std::log2(x) : acc;
        };
        // We use `std::abs` rather than negation because we never want -0 as a result.
        return std::abs(std::accumulate(first, last, zero, lambda));
    }

    template <typename ContainerT, typename ValueT = typename ContainerT::value_type>
    std::enable_if_t<std::is_floating_point_v<ValueT>, ValueT>
    entropy(const ContainerT& container)
    {
        return entropy(std::begin(container), std::end(container));
    }

    struct gaussian final
    {
        gaussian() noexcept = default;

        gaussian(const double mu, const double sigma) : _mu{mu}, _var{sigma * sigma}
        {
            assert(sigma > 0.0);
        }

        double operator()(const double x) const noexcept
        {
            const auto dx = x - _mu;
            return 1.0 / std::sqrt(2.0 * M_PI * _var) * std::exp(-(dx * dx) / (2.0 * _var));
        }

    private:

        double _mu{0.0};
        double _var{1.0};

    };  // struct gaussian

}  // namespace msc

#endif  // !defined(MSC_STOCHASTIC_HXX)
