/* -*- coding:utf-8; mode:c++; -*- */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <cstddef>
#include <stdexcept>
#include <utility>

#include <ogdf/basic/Graph.h>
#include <ogdf/basic/GraphAttributes.h>

#include "cli.hxx"
#include "data_analysis.hxx"
#include "io.hxx"
#include "json.hxx"
#include "meta.hxx"
#include "ogdf_fix.hxx"
#include "point.hxx"
#include "rdf.hxx"

#define PROGRAM_NAME "rdf-local"

namespace /*anonymous*/
{

    double get_longest_path(const msc::ogdf_node_array_2d<double>& matrix, const ogdf::Graph& graph) noexcept
    {
        auto longest = 0.0;
        for (auto v1 = graph.firstNode(); v1 != nullptr; v1 = v1->succ()) {
            for (auto v2 = v1->succ(); v2 != nullptr; v2 = v2->succ()) {
                const auto v1tov2 = matrix[v1][v2];
                // OGDF inserts nonsensical large but finite values into the matrix for unreachable nodes.
                if ((v1tov2 <= graph.numberOfNodes()) && (v1tov2 > longest)) {
                    longest = v1tov2;
                }
            }
        }
        return longest;
    }

    struct application final
    {
        msc::cli_parameters_property parameters{};
        void operator()() const;
    };

    msc::json_object basic_info()
    {
        auto info = msc::json_object{};
        info["producer"] = PROGRAM_NAME;
        return info;
    }

    void application::operator()() const
    {
        if (this->parameters.iterations() > 1) {
            throw std::invalid_argument{"The '--width' and '--bins' options must be passed at most once"};
        }
        const auto [graph, attrs] = msc::load_layout(this->parameters.input);
        const auto matrix = msc::get_pairwise_shortest_paths(*graph);
        const auto longestpath = get_longest_path(*matrix, *graph);
        const auto maxdist = abs(msc::get_bounding_box_size(*attrs));
        auto distances = msc::local_pairwise_distances{*attrs, *matrix, NAN};
        auto info = basic_info();
        auto subinfos = msc::json_array{};
        auto analyzer = msc::data_analyzer{this->parameters.kernel};
        auto entropies = msc::initialize_entropies();
        analyzer.set_range(0.0, maxdist);
        analyzer.set_width(msc::get_item(this->parameters.width, 0));
        analyzer.set_bins(msc::get_item(this->parameters.bins, 0));
        analyzer.set_points(this->parameters.points);
        for (std::size_t i = 0; true; ++i) {
            const auto vicinity = (std::size_t{1} << i);
            auto subinfo = msc::json_object{};
            subinfo["vicinity"] = msc::json_size{vicinity};
            distances.set_limit(vicinity);
            analyzer.set_output(this->parameters.expand_filename(i));
            if (analyzer.analyze_oknodo(distances, subinfo, subinfo)) {
                msc::append_entropy(entropies, subinfo, "vicinity");
                subinfos.push_back(std::move(subinfo));
            }
            if (vicinity >= longestpath) { break; }
        }
        if (subinfos.empty()) {
            throw std::runtime_error{"Not enough data for a statistical analysis"};
        }
        info["data"] = std::move(subinfos);
        msc::assign_entropy_regression(entropies, info);
        info["diameter"] = msc::json_real{longestpath};
        msc::print_meta(info, this->parameters.meta);
    }

}  // namespace /*anonymous*/

int main(const int argc, const char *const *const argv)
{
    auto app = msc::command_line_interface<application>{PROGRAM_NAME};
    app.help.push_back(
        "Computes the local radial distribution function (RDF) for a graph layout.  Local means that only pairs of"
        " nodes will be considered for which the shortest path is at most 2^l for locality l which will be incremented"
        " starting from 0 up to a value where all pairs of nodes are covered."
    );
    app.help.push_back(msc::helptext_file_name_expansion());
    return app(argc, argv);
}
