/* -*- coding:utf-8; mode:c++; -*- */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#define MSC_RUN_ALL_UNIT_TESTS_IN_MAIN

#include "stochastic.hxx"

#include <algorithm>
#include <array>
#include <cstddef>
#include <random>
#include <utility>
#include <vector>

#include "unittest.hxx"

namespace /*anonymous*/
{

    MSC_AUTO_TEST_CASE(square)
    {
        static_assert(0 == msc::square(0));
        static_assert(1.0 == msc::square(1.0));
        static_assert(4.0 == msc::square(2.0));
        static_assert(529 == msc::square(23));
    }

    MSC_AUTO_TEST_CASE(mean_stdev_1st)
    {
        constexpr auto tol = 1.0e-15;
        const std::array<double, 3> numbers = {{42.0, 42.0, 42.0}};
        const auto [mean, stdev] = msc::mean_stdev(numbers);
        MSC_REQUIRE_CLOSE(tol, 42.0, mean);
        MSC_REQUIRE_CLOSE(tol, 0.0, stdev);
    }

    MSC_AUTO_TEST_CASE(mean_stdev_2nd)
    {
        constexpr auto tol = 1.0e-10;
        constexpr auto n = std::size_t{100};
        const auto numbers = [](){
            auto vec = std::vector<double>(n);
            std::iota(std::begin(vec), std::end(vec), 1.0);
            return vec;
        }();
        MSC_REQUIRE_EQ(1.0, numbers.front());
        MSC_REQUIRE_EQ(100.0, numbers.back());
        const auto [mean, stdev] = msc::mean_stdev(numbers);
        MSC_REQUIRE_CLOSE(tol, 50.500000000000000, mean);
        MSC_REQUIRE_CLOSE(tol, 29.011491975882000, stdev);
    }

    MSC_AUTO_TEST_CASE(mean_stdev_3rd)
    {
        constexpr auto reltol = 1.0e-3;
        constexpr auto n = std::size_t{600000};
        auto rndeng = std::mt19937{};
        auto metadist = std::uniform_real_distribution<double>{1.0, 2.0};
        auto rnddst = std::normal_distribution<double>{100.0 * metadist(rndeng), 3.0 * metadist(rndeng)};
        auto numbers = std::vector<double>(n);
        std::generate(std::begin(numbers), std::end(numbers), [&rndeng, &rnddst](){ return rnddst(rndeng); });
        const auto [mean, stdev] = msc::mean_stdev(numbers);
        MSC_REQUIRE_CLOSE(reltol, 1.0, rnddst.mean() / mean);
        MSC_REQUIRE_CLOSE(reltol, 1.0, rnddst.stddev() / stdev);
    }

    MSC_AUTO_TEST_CASE(min_max_1)
    {
        const auto numbers = std::initializer_list<int>{1};
        const auto [min, max] = msc::min_max(numbers);
        MSC_REQUIRE_EQ(1, min);
        MSC_REQUIRE_EQ(1, max);
    }

    MSC_AUTO_TEST_CASE(min_max_2)
    {
        const auto numbers = std::initializer_list<int>{7, 12};
        const auto [min, max] = msc::min_max(numbers);
        MSC_REQUIRE_EQ(7, min);
        MSC_REQUIRE_EQ(12, max);
    }

    MSC_AUTO_TEST_CASE(min_max_3)
    {
        const auto numbers = std::initializer_list<int>{1, 2, 3};
        const auto [min, max] = msc::min_max(numbers);
        MSC_REQUIRE_EQ(1, min);
        MSC_REQUIRE_EQ(3, max);
    }

    MSC_AUTO_TEST_CASE(min_max_n)
    {
        constexpr auto n = std::size_t{1000};
        const auto numbers = [](){
            auto vec = std::vector<double>(n);
            auto rndeng = std::mt19937{};
            auto rnddst = std::uniform_real_distribution<double>{14.0, 92.0};
            std::generate(std::begin(vec), std::end(vec), [&rndeng, &rnddst](){ return rnddst(rndeng); });
            return vec;
        }();
        const auto [min, max] = msc::min_max(numbers);
        MSC_REQUIRE_LE(min, max);
        MSC_REQUIRE(std::all_of(std::begin(numbers), std::end(numbers), [min](const double x){ return (min <= x); }));
        MSC_REQUIRE(std::all_of(std::begin(numbers), std::end(numbers), [max](const double x){ return (max >= x); }));
        MSC_REQUIRE(std::any_of(std::begin(numbers), std::end(numbers), [min](const double x){ return (min == x); }));
        MSC_REQUIRE(std::any_of(std::begin(numbers), std::end(numbers), [max](const double x){ return (max == x); }));
    }

}  // namespace /*anonymous*/
