/* -*- coding:utf-8; mode:c++; -*- */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#define MSC_RUN_ALL_UNIT_TESTS_IN_MAIN

#include "useful.hxx"

#include <climits>
#include <string>
#include <string_view>

#include "unittest.hxx"

namespace /*anonymous*/
{

    MSC_AUTO_TEST_CASE(optional_cast_value)
    {
        const std::optional<double> src = 13.7;
        const std::optional<int> dst = msc::optional_cast<int>(src);
        MSC_REQUIRE(dst.has_value());
        MSC_REQUIRE_EQ(13, dst.value());
    }

    MSC_AUTO_TEST_CASE(optional_cast_empty)
    {
        const std::optional<double> src = std::nullopt;
        const std::optional<int> dst = msc::optional_cast<int>(src);
        MSC_REQUIRE(!dst.has_value());
    }

    MSC_AUTO_TEST_CASE(share_pair_00)
    {
        using shared_pair = std::pair<std::shared_ptr<double>, std::shared_ptr<char>>;
        using unique_pair = std::pair<std::unique_ptr<double>, std::unique_ptr<char>>;
        unique_pair unique = {nullptr, nullptr};
        shared_pair shared = msc::share_pair(std::move(unique));
        MSC_REQUIRE_EQ(nullptr, unique.first.get());
        MSC_REQUIRE_EQ(nullptr, unique.second.get());
        MSC_REQUIRE_EQ(nullptr, shared.first.get());
        MSC_REQUIRE_EQ(nullptr, shared.second.get());
    }

    MSC_AUTO_TEST_CASE(share_pair_01)
    {
        using shared_pair = std::pair<std::shared_ptr<double>, std::shared_ptr<char>>;
        using unique_pair = std::pair<std::unique_ptr<double>, std::unique_ptr<char>>;
        unique_pair unique = {nullptr, std::make_unique<char>('A')};
        shared_pair shared = msc::share_pair(std::move(unique));
        MSC_REQUIRE_EQ(nullptr, unique.first.get());
        MSC_REQUIRE_EQ(nullptr, unique.second.get());
        MSC_REQUIRE_EQ(nullptr, shared.first.get());
        MSC_REQUIRE_NE(nullptr, shared.second.get());
        MSC_REQUIRE_EQ('A', *shared.second);
    }

    MSC_AUTO_TEST_CASE(share_pair_10)
    {
        using shared_pair = std::pair<std::shared_ptr<double>, std::shared_ptr<char>>;
        using unique_pair = std::pair<std::unique_ptr<double>, std::unique_ptr<char>>;
        unique_pair unique = {std::make_unique<double>(1.5), nullptr};
        shared_pair shared = msc::share_pair(std::move(unique));
        MSC_REQUIRE_EQ(nullptr, unique.first.get());
        MSC_REQUIRE_EQ(nullptr, unique.second.get());
        MSC_REQUIRE_NE(nullptr, shared.first.get());
        MSC_REQUIRE_EQ(nullptr, shared.second.get());
        MSC_REQUIRE_EQ(1.5, *shared.first);
    }

    MSC_AUTO_TEST_CASE(share_pair_11)
    {
        using shared_pair = std::pair<std::shared_ptr<double>, std::shared_ptr<char>>;
        using unique_pair = std::pair<std::unique_ptr<double>, std::unique_ptr<char>>;
        unique_pair unique = {std::make_unique<double>(1.5), std::make_unique<char>('A')};
        shared_pair shared = msc::share_pair(std::move(unique));
        MSC_REQUIRE_EQ(nullptr, unique.first.get());
        MSC_REQUIRE_EQ(nullptr, unique.second.get());
        MSC_REQUIRE_NE(nullptr, shared.first.get());
        MSC_REQUIRE_NE(nullptr, shared.second.get());
        MSC_REQUIRE_EQ(1.5, *shared.first);
        MSC_REQUIRE_EQ('A', *shared.second);
    }

    MSC_AUTO_TEST_CASE(normalize_constant_name)
    {
        MSC_REQUIRE_EQ("", msc::normalize_constant_name(""));
        MSC_REQUIRE_EQ("alpha", msc::normalize_constant_name("alpha"));
        MSC_REQUIRE_EQ("", msc::normalize_constant_name("  \t \t\t"));
        MSC_REQUIRE_EQ("alpha", msc::normalize_constant_name("   alpha"));
        MSC_REQUIRE_EQ("alpha", msc::normalize_constant_name("alpha   "));
        MSC_REQUIRE_EQ("alpha", msc::normalize_constant_name(" \t  alpha  \t\t"));
        MSC_REQUIRE_EQ("alpha", msc::normalize_constant_name("Alpha"));
        MSC_REQUIRE_EQ("alpha", msc::normalize_constant_name("ALPHA"));
        MSC_REQUIRE_EQ("a   a", msc::normalize_constant_name("\tA   a     "));
    }

    MSC_AUTO_TEST_CASE(reject_invalid_enumeration_enum)
    {
        using namespace std::string_view_literals;
        enum class moods { happy = 1, angry = 2, sad = 3 };
        try {
            msc::reject_invalid_enumeration(moods{}, "moods");
            MSC_FAIL("No exception was thrown");
        } catch (const std::invalid_argument& e) {
            MSC_REQUIRE_EQ("0 is not a valid constant for an enumerator of type 'moods'"sv, std::string_view{e.what()});
        }
    }

    MSC_AUTO_TEST_CASE(reject_invalid_enumeration_int)
    {
        using namespace std::string_view_literals;
        try {
            msc::reject_invalid_enumeration(42, "moods");
            MSC_FAIL("No exception was thrown");
        } catch (const std::invalid_argument& e) {
            MSC_REQUIRE_EQ(
                "42 is not a valid constant for an enumerator of type 'moods'"sv,
                std::string_view{e.what()}
            );
        }
    }

    MSC_AUTO_TEST_CASE(reject_invalid_enumeration_str)
    {
        using namespace std::string_view_literals;
        try {
            msc::reject_invalid_enumeration("dizzy", "moods");
            MSC_FAIL("No exception was thrown");
        } catch (const std::invalid_argument& e) {
            MSC_REQUIRE_EQ(
                "'dizzy' is not a valid name for an enumerator of type 'moods'"sv,
                std::string_view{e.what()}
            );
        }
    }

    MSC_AUTO_TEST_CASE(parse_decimal_number)
    {
        MSC_REQUIRE_EQ(0, msc::parse_decimal_number("0"));
        MSC_REQUIRE_EQ(1, msc::parse_decimal_number("1"));
        MSC_REQUIRE_EQ(42, msc::parse_decimal_number("42"));
        MSC_REQUIRE_EQ(INT_MAX, msc::parse_decimal_number(std::to_string(INT_MAX)));
        MSC_REQUIRE_EQ(std::nullopt, msc::parse_decimal_number(""));
        MSC_REQUIRE_EQ(std::nullopt, msc::parse_decimal_number("-1"));
        MSC_REQUIRE_EQ(std::nullopt, msc::parse_decimal_number("Holger"));
        MSC_REQUIRE_EQ(std::nullopt, msc::parse_decimal_number("0x20"));
        MSC_REQUIRE_EQ(std::nullopt, msc::parse_decimal_number("2f"));
        MSC_REQUIRE_EQ(std::nullopt, msc::parse_decimal_number(" 4"));
        MSC_REQUIRE_EQ(std::nullopt, msc::parse_decimal_number("4 "));
        MSC_REQUIRE_EQ(std::nullopt, msc::parse_decimal_number(" "));
        MSC_REQUIRE_EQ(std::nullopt, msc::parse_decimal_number("99999999999999999999999999999999999999999999999"));
    }

}  // namespace /*anonymous*/
